from __future__ import annotations

import contextlib
import random
from pydantic import BaseModel, ConfigDict, Field
from typing import List, Optional

import numpy as np
import time
from functools import cached_property

from prodsys.adapters import adapter
from prodsys.simulation import sim, logger
from prodsys.factories import (
    link_transport_process_updater,
    auxiliary_factory,
    state_factory,
    time_model_factory,
    process_factory,
    queue_factory,
    resource_factory,
    product_factory,
    sink_factory,
    source_factory,
    node_factory,
)

from prodsys.util.post_processing import PostProcessor

from prodsys.util import kpi_visualization, util
from prodsys.models import performance_data

VERBOSE = 1

def run_simulation(adapter_object: adapter.ProductionSystemAdapter, run_length: int) -> Runner:
    """
    Runs the simulation for the given adapter and run length.

    Args:
        adapter_object (adapter.ProductionSystemAdapter): Adapter containing the production system to simulate.
        run_length (int): Length of the simulation run.

    Returns:
        Runner: The runner object after simulation.
    """
    runner_object = Runner(adapter=adapter_object)
    runner_object.initialize_simulation()
    runner_object.run(run_length)
    return runner_object


@contextlib.contextmanager
def temp_seed(seed: int):
    """
    Context manager for temporarily setting the seed of the random number generators. Is necessary when optimizing with another random seed but still wanting to use the same seed for the simulation.

    Args:
        seed (int): The seed to set for the simulation run.
    """
    np_state = np.random.get_state()
    p_state = random.getstate()
    np.random.seed(seed)
    random.seed(seed)
    try:
        yield
    finally:
        np.random.set_state(np_state)
        random.setstate(p_state)


class Runner:
    """
    Class to represent the simulation runner. It allows to run the simulation based on a provided adapter.

    Args:
        adapter (adapter.ProductionSystemAdapter): The adapter containing the production system to simulate.

    Attributes:
        adapter (adapter.ProductionSystemAdapter): The adapter containing the production system to simulate.
        env (sim.Environment): The environment to run the simulation in.
        time_model_factory (time_model_factory.TimeModelFactory): The time model factory to create the time models.
        state_factory (state_factory.StateFactory): The state factory to create the states.
        process_factory (process_factory.ProcessFactory): The process factory to create the processes.
        queue_factory (queue_factory.QueueFactory): The queue factory to create the queues.
        resource_factory (resource_factory.ResourceFactory): The resource factory to create the resources.
        sink_factory (sink_factory.SinkFactory): The sink factory to create the sinks.
        source_factory (source_factory.SourceFactory): The source factory to create the sources.
        product_factory (product_factory.ProductFactory): The product factory to create the products.
        event_logger (logger.Logger): The event logger to log the events.
        time_stamp (str): The time stamp of the simulation run.
        post_processor (post_processing.PostProcessor): The post processor to process the simulation results.
    """

    def __init__(self, adapter: adapter.ProductionSystemAdapter):
        self.adapter = adapter
        self.env = sim.Environment(seed=self.adapter.seed)
        self.time_model_factory: time_model_factory.TimeModelFactory = None
        self.state_factory: state_factory.StateFactory = None
        self.process_factory: process_factory.ProcessFactory = None
        self.queue_factory: queue_factory.QueueFactory = None
        self.auxiliary_factory: auxiliary_factory.AuxiliaryFactory = None
        self.resource_factory: resource_factory.ResourceFactory = None
        self.node_factory: node_factory.NodeFactory = None
        self.sink_factory: sink_factory.SinkFactory = None
        self.source_factory: source_factory.SourceFactory = None
        self.product_factory: product_factory.ProductFactory = None
        self.event_logger: logger.Logger = None
        self.time_stamp: str = ""
        self.post_processor: PostProcessor = None


    def initialize_simulation(self):
        """
        Initializes the simulation by creating the factories and all simulation objects. Needs to be done before running the simulation.
        """
        self.adapter.validate_configuration()
        with temp_seed(self.adapter.seed):

            self.time_model_factory = time_model_factory.TimeModelFactory()
            self.time_model_factory.create_time_models(self.adapter)

            self.env = sim.Environment(seed=self.adapter.seed)

            self.state_factory = state_factory.StateFactory(
                env=self.env, time_model_factory=self.time_model_factory
            )
            self.state_factory.create_states(self.adapter)

            self.process_factory = process_factory.ProcessFactory(
                time_model_factory=self.time_model_factory
            )
            self.process_factory.create_processes(self.adapter)

            self.queue_factory = queue_factory.QueueFactory(env=self.env)
            self.queue_factory.create_queues(self.adapter)

            self.resource_factory = resource_factory.ResourceFactory(
                env=self.env,
                state_factory=self.state_factory,
                queue_factory=self.queue_factory,
                process_factory= self.process_factory
            )
            self.resource_factory.create_resources(self.adapter)

            self.node_factory = node_factory.NodeFactory(
                env=self.env)
            self.node_factory.create_nodes(self.adapter)

            self.product_factory = product_factory.ProductFactory(
                env=self.env, 
                process_factory=self.process_factory,
            )

            self.sink_factory = sink_factory.SinkFactory(
                env=self.env,
                product_factory=self.product_factory,
                queue_factory=self.queue_factory,
            )

            self.sink_factory.create_sinks(self.adapter)

            self.auxiliary_factory = auxiliary_factory.AuxiliaryFactory(
                env=self.env,
                process_factory=self.process_factory,
                queue_factory=self.queue_factory,
                resource_factory=self.resource_factory,
                sink_factory= self.sink_factory
            )

            self.product_factory = product_factory.ProductFactory(
                env=self.env, 
                process_factory=self.process_factory,
            )

            self.sink_factory = sink_factory.SinkFactory(
                env=self.env,
                product_factory=self.product_factory,
                queue_factory=self.queue_factory,
            )

            self.sink_factory.create_sinks(self.adapter)

            self.event_logger = logger.EventLogger()
            self.event_logger.observe_resource_states(self.resource_factory)

            self.product_factory.event_logger = self.event_logger
            self.auxiliary_factory.event_logger = self.event_logger
            self.auxiliary_factory.create_auxiliary(self.adapter)


            self.source_factory = source_factory.SourceFactory(
                env=self.env,
                product_factory=self.product_factory,
                time_model_factory=self.time_model_factory,
                queue_factory=self.queue_factory,
                resource_factory=self.resource_factory,
                auxiliary_factory=self.auxiliary_factory,
                sink_factory=self.sink_factory,
            )
            self.source_factory.create_sources(self.adapter)

            link_transport_process_updater_instance = link_transport_process_updater.LinkTransportProcessUpdater(
                process_factory=self.process_factory,
                source_factory=self.source_factory,
                sink_factory=self.sink_factory,
                resource_factory=self.resource_factory,
                node_factory=self.node_factory,
            )
            link_transport_process_updater_instance.update_links_with_objects()
            
            self.auxiliary_factory.place_auxiliaries_in_queues()
            self.resource_factory.start_resources()
            self.source_factory.start_sources()

    def run(self, time_range: int):
        """
        Runs the simulation for the given time range.

        Args:
            time_range (int): The time range to run the simulation for.
        """

        t_0 = time.perf_counter()

        self.env.run(time_range)

        t_1 = time.perf_counter()
        self.time_stamp = time.strftime("%Y%m%d-%H%M%S")

    def get_post_processor(self) -> PostProcessor:
        """
        Returns the post processor to process the simulation results.

        Returns:
            post_processing.PostProcessor: The post processor to process the simulation results.
        """
        if not self.post_processor:
            self.post_processor = PostProcessor(df_raw=self.event_logger.get_data_as_dataframe())
        return self.post_processor


    def print_results(self):
        """
        Prints the aggregated simulation results, comprising the average throughput, WIP, throughput time and the time per state of the resources.
        """
        p = self.get_post_processor()
        kpi_visualization.print_aggregated_data(p)

    def plot_results(self):
        """
        Plots the aggregated simulation results, comprising the throughput time over time, WIP over time, throughput time distribution and the time per state of the resources.
        """
        p = self.get_post_processor()
        kpi_visualization.plot_throughput_time_over_time(p)
        kpi_visualization.plot_WIP(p)
        if self.adapter.auxiliary_data:
            kpi_visualization.plot_auxiliary_WIP(p)
        # kpi_visualization.plot_WIP_per_resource(p)
        kpi_visualization.plot_throughput_time_distribution(p)
        kpi_visualization.plot_time_per_state_of_resources(p)

    def plot_results_executive(self):
        """
        Plots the aggregated simulation results, comprising the throughput time over time, WIP over time, throughput time distribution and the time per state of the resources.
        """
        p = self.get_post_processor()
        kpi_visualization.plot_boxplot_resource_utilization(p)
        kpi_visualization.plot_line_balance_kpis(p)
        kpi_visualization.plot_production_flow_rate_per_product(p)
        transport_resource_ids = [resource_data.ID for resource_data in adapter.get_transport_resources(self.adapter)]
        kpi_visualization.plot_transport_utilization_over_time(p, transport_resource_ids)
        kpi_visualization.plot_util_WIP_resource(p)
        kpi_visualization.plot_oee(p)


    def get_event_data_of_simulation(self) -> List[performance_data.Event]:
        """
        Returns the event data of the simulation.

        Returns:
            List[performance_data.Event]: The event data of the simulation.
        """
        p = self.get_post_processor()
        df_raw=self.event_logger.get_data_as_dataframe()
        events = []
        df_raw["Expected End Time"] = df_raw["Expected End Time"].fillna(value=-1)
        df_raw["Target location"] = df_raw["Target location"].fillna(value="")
        df_raw["Product"] = df_raw["Product"].fillna(value="")
        for index, row in df_raw.iterrows():
            
            events.append(
                performance_data.Event(
                    time=row["Time"],
                    resource=row["Resource"],
                    state=row["State"],
                    state_type=row["State Type"],
                    activity=row["Activity"],
                    product=row["Product"],
                    expected_end_time=row["Expected End Time"],
                    target_location=row["Target location"],
                )
            )
        return events
    
    def get_performance_data(self) -> performance_data.Performance:
        """
        Returns the performance data of the simulation.

        Returns:
            performance_data.Performance: The performance data of the simulation.
        """
        p = self.get_post_processor()
        kpis = []
        kpis += p.WIP_KPIs
        kpis += p.throughput_and_output_KPIs
        kpis += p.aggregated_throughput_time_KPIs
        kpis += p.machine_state_KPIS
        event_data = self.get_event_data_of_simulation()
        return performance_data.Performance(kpis=kpis, event_log=event_data)
    
    def get_aggregated_data_simulation_results(self) -> dict:
        """
        Returns the aggregated simulation results.

        Returns:
            dict: The aggregated simulation results.
        """
        p = PostProcessor(df_raw=self.event_logger.get_data_as_dataframe())
        return p.get_aggregated_data()

    def save_results_as_csv(self, save_folder="data"):
        """
        Saves the simulation results as .csv-file marked with the time_stamp of simulation and the adapter ID if available.

        Args:
            save_folder (str, optional): The folder to save the results to. Defaults to "data".
        """
        util.prepare_save_folder(save_folder + "/")
        save_name = ""
        if self.adapter.ID:
            save_name = f"{self.adapter.ID}_"
        save_name += self.time_stamp
        self.event_logger.log_data_to_csv(filepath=f"{save_folder}/{save_name}.csv")

    def save_results_as_json(self, save_folder="data"):
        """
        Saves the simulation results as .json-file marked with the time_stamp of simulation and the adapter ID if available.

        Args:
            save_folder (str, optional): The folder to save the results to. Defaults to "data".
        """
        util.prepare_save_folder(save_folder + "/")
        save_name = ""
        if self.adapter.ID:
            save_name = f"{self.adapter.ID}_"
        save_name += self.time_stamp
        self.event_logger.log_data_to_json(filepath=f"{save_folder}/{save_name}.json")




import json
import time
from copy import deepcopy

from pydantic import BaseModel, ConfigDict
import logging

from prodsys.optimization.optimization import evaluate
from prodsys.optimization.adapter_manipulation import mutation
from prodsys.optimization.optimization import check_valid_configuration
from prodsys.optimization.util import document_individual

logger = logging.getLogger(__name__)

from simanneal import Annealer

from prodsys.simulation import sim
from prodsys import adapters
from prodsys.optimization.util import (
    get_weights,
    check_breakdown_states_available,
    create_default_breakdown_states,
)
from prodsys.util.util import set_seed

sim.VERBOSE = 0


class ProductionSystemOptimization(Annealer):
    def __init__(
        self,
        base_configuration: adapters.ProductionSystemAdapter,
        save_folder: str,
        performances: dict,
        solutions_dict: dict,
        start: float,
        weights: tuple,
        number_of_seeds: int = 1,
        initial_solution: adapters.ProductionSystemAdapter = None,
        full_save: bool = False,
    ):
        super().__init__(initial_solution, None)
        self.save_folder = save_folder
        self.base_configuration = base_configuration
        self.performances = performances
        self.solution_dict = solutions_dict
        self.start = start
        self.weights = weights
        self.number_of_seeds = number_of_seeds
        self.full_save = full_save

    def move(self):
        while True:
            configuration = mutation(individual=[deepcopy(self.state)])[0][0]
            if check_valid_configuration(
                configuration=configuration,
                base_configuration=self.base_configuration,
            ):
                self.state = configuration
                break

    def energy(self):
        values = evaluate(
            base_scenario=self.base_configuration,
            performances=self.performances,
            solution_dict=self.solution_dict,
            number_of_seeds=self.number_of_seeds,
            full_save_folder_file_path=self.save_folder if self.full_save else "",
            individual=[self.state],
        )

        performance = sum(
            [value * weight for value, weight in zip(values, self.weights)]
        )
        # print("\n\t########## Evaluted ind", self.counter, "for value:", performance)
        counter = len(self.performances["0"]) - 1
        document_individual(self.solution_dict, self.save_folder, [self.state])
        self.performances["0"][self.state.ID] = {
            "agg_fitness": performance,
            "fitness": [float(value) for value in values],
            "time_stamp": time.perf_counter() - self.start,
            "hash": self.state.hash(),
        }
        with open(f"{self.save_folder}/optimization_results.json", "w") as json_file:
            json.dump(self.performances, json_file)

        return performance


class SimulatedAnnealingHyperparameters(BaseModel):
    """
    Hyperparameters to perform a configuration optimization with simulated annealing.
    Args:
        seed (int): Seed for random number generator
        Tmax (int): Maximum temperature
        Tmin (int): Minimum temperature
        steps (int): Number of steps
        updates (int): Number of updates
    """

    seed: int = 0
    Tmax: int = 10000
    Tmin: int = 1
    steps: int = 4000
    updates: int = 300
    number_of_seeds: int = 1

    model_config = ConfigDict(
        json_schema_extra={
            "examples": [
                {
                    "seed": 0,
                    "Tmax": 10000,
                    "Tmin": 1,
                    "steps": 4000,
                    "updates": 300,
                    "number_of_seeds": 1,
                },
            ]
        }
    )


def run_simulated_annealing(
    save_folder: str,
    base_configuration_file_path: str,
    scenario_file_path: str,
    full_save: bool,
    seed: int,
    Tmax: int,
    Tmin: int,
    steps: int,
    updates: int,
    number_of_seeds: int,
    initial_solution_file_path: str = "",
):
    """
    Run a simulated annealing algorithm for configuration optimization.

    Args:
        save_folder (str): Folder to save the results in.
        base_configuration_file_path (str): File path of the serialized base configuration (`prodsys.adapters.JsonProductionSystemAdapter`)
        scenario_file_path (str): File path of the serialized scenario (`prodsys.models.scenario_data.ScenarioData`)
        full_save (bool): Save the full results of the optimization.
        seed (int): Random seed for optimization.
        Tmax (int): Maximum temperature
        Tmin (int): Minimum temperature
        steps (int): Steps for annealing
        updates (int): Number of updates
        number_of_seeds (int): Number of seeds for optimization
        initial_solution_file_path (str, optional): File path to an initial solution. Defaults to "".
    """
    base_configuration = adapters.JsonProductionSystemAdapter()
    base_configuration.read_data(base_configuration_file_path, scenario_file_path)
    if not base_configuration.ID:
        base_configuration.ID = "base_configuration"

    if initial_solution_file_path:
        initial_solution = adapters.JsonProductionSystemAdapter()
        initial_solution.read_data(initial_solution_file_path, scenario_file_path)
        if not initial_solution.ID:
            initial_solution.ID = "initial_solution"
    else:
        initial_solution = base_configuration.model_copy(deep=True)

    hyper_parameters = SimulatedAnnealingHyperparameters(
        seed=seed,
        Tmax=Tmax,
        Tmin=Tmin,
        steps=steps,
        updates=updates,
        number_of_seeds=number_of_seeds,
    )

    simulated_annealing_optimization(
        base_configuration=base_configuration,
        hyper_parameters=hyper_parameters,
        save_folder=save_folder,
        initial_solution=initial_solution,
        full_save=full_save,
    )


def simulated_annealing_optimization(
    base_configuration: adapters.ProductionSystemAdapter,
    hyper_parameters: SimulatedAnnealingHyperparameters,
    save_folder: str = "results",
    initial_solution: adapters.ProductionSystemAdapter = None,
    full_save: bool = False,
):
    """
    Optimize a production system configuration using simulated anealing.

    Args:
        base_configuration (adapters.ProductionSystemAdapter): production system to optimize.
        hyper_parameters (SimulatedAnnealingHyperparameters): Hyperparameters for simulated annealing.
        save_folder (str): Folder to save the results in. Defaults to "results".
        initial_solution (adapters.ProductionSystemAdapter, optional): Initial solution for optimization. Defaults to None.
    """
    adapters.ProductionSystemAdapter.model_config["validate_assignment"] = False
    if not adapters.check_for_clean_compound_processes(base_configuration):
        logger.warning(
            "Both compound processes and normal processes are used. This may lead to unexpected results."
        )
    if not check_breakdown_states_available(base_configuration):
        create_default_breakdown_states(base_configuration)
    if not initial_solution:
        initial_solution = base_configuration.model_copy(deep=True)

    set_seed(hyper_parameters.seed)

    weights = get_weights(base_configuration, "min")

    solution_dict = {"current_generation": "0", "hashes": {}}
    performances = {}
    performances["0"] = {}
    start = time.perf_counter()

    pso = ProductionSystemOptimization(
        base_configuration=base_configuration,
        save_folder=save_folder,
        performances=performances,
        solutions_dict=solution_dict,
        start=start,
        weights=weights,
        number_of_seeds=hyper_parameters.number_of_seeds,
        initial_solution=initial_solution,
        full_save=full_save,
    )

    pso.Tmax = hyper_parameters.Tmax
    pso.Tmin = hyper_parameters.Tmin
    pso.steps = hyper_parameters.steps
    pso.updates = hyper_parameters.updates

    internary, performance = pso.anneal()


def optimize_configuration(
    base_configuration_file_path: str,
    scenario_file_path: str,
    save_folder: str,
    hyper_parameters: SimulatedAnnealingHyperparameters,
    full_save: bool = False,
):
    """
    Optimize a configuration with simulated annealing.

    Args:
        base_configuration_file_path (str): File path of the serialized base configuration (`prodsys.adapters.JsonProductionSystemAdapter`)
        scenario_file_path (str): File path of the serialized scenario (`prodsys.models.scenario_data.ScenarioData`)
        save_folder (str): Folder to save the results in.
        hyper_parameters (SimulatedAnnealingHyperparameters): Hyperparameters to perform a configuration optimization with simulated annealing.
        full_save (bool, optional): Save the full results of the optimization. Defaults to False.
    """
    run_simulated_annealing(
        save_folder=save_folder,
        base_configuration_file_path=base_configuration_file_path,
        scenario_file_path=scenario_file_path,
        seed=hyper_parameters.seed,
        Tmax=hyper_parameters.Tmax,
        Tmin=hyper_parameters.Tmin,
        steps=hyper_parameters.steps,
        updates=hyper_parameters.updates,
        number_of_seeds=hyper_parameters.number_of_seeds,
        full_save=full_save,
    )



from typing import List, Optional, Union


from pydantic import Field, PrivateAttr
from pydantic.dataclasses import dataclass

from prodsys.util import util
from prodsys.util import runner
from prodsys.adapters import adapter, json_adapter

from prodsys.express import (
    core, node, resources, source, time_model, state, process, product, sink
)


def remove_duplicate_items(
    items: List[
        Union[
            resources.Resource,
            source.Source,
            sink.Sink,
            product.Product,
            state.STATE_UNION,
            process.PROCESS_UNION,
            time_model.TIME_MODEL_UNION,
        ]
    ]
) -> List[
    Union[
        resources.Resource,
        source.Source,
        sink.Sink,
        product.Product,
        state.STATE_UNION,
        process.PROCESS_UNION,
        time_model.TIME_MODEL_UNION,
    ]
]:
    """
    Removes duplicate items from a list of items.
    """
    id_set = set()
    filtered_items = []
    for item in items:
        if item.ID in id_set:
            continue
        id_set.add(item.ID)
        filtered_items.append(item)
    return filtered_items

@dataclass
class ProductionSystem(core.ExpressObject):
    """
    Class that represents a production system. A production system containts of resources, products, sources and sinks.
    It is the `prodsys.express` equivalent to the 'ProductionSystemAdapter' of the `prodsys.adapters` module and
    can be converted to this data object. In contrast to the adapter,
    this class nests the objects in a tree structure, which makes it easier to work with when instantiating
    a production system, but more complicated when reviewing the data itself.

    Args:
        resources (List[resources.Resource]): Resources of the production system.
        sources (List[source.Source]): Sources of the production system.
        sinks (List[sink.Sink]): Sinks of the production system.
    """
    resources: List[Union[resources.ProductionResource, resources.TransportResource]]
    sources: List[source.Source]
    sinks: List[sink.Sink]

    def __post_init__(self):
        self._runner: Optional[runner.Runner] = None

    def to_model(self) -> adapter.ProductionSystemAdapter:
        """
        Converts the `prodsys.express` object to a data object from `prodsys.models`.

        Returns:
            prodsys.adapters.Adapter: An instance of the data object.
        """
        products = [source.product for source in self.sources] + [
            sink.product for sink in self.sinks
        ]
        products = remove_duplicate_items(products)
        auxiliaries = list(util.flatten_object([product.auxiliaries for product in products]))
        auxiliaries = remove_duplicate_items(auxiliaries)
        processes = list(
            util.flatten_object(
                [product.processes for product in products]+
                [product.transport_process for product in products]
                + [resource.processes for resource in self.resources]
            )
        )
        processes = remove_duplicate_items(processes)

        nodes = []
        for process_instance in processes:
            if not isinstance(process_instance, process.LinkTransportProcess):
                continue
            for link in process_instance.links:
                for link_element in link:
                    if isinstance(link_element, node.Node):
                        nodes.append(link_element)
        nodes = remove_duplicate_items(nodes)
        auxiliary_storages = list(util.flatten_object(auxiliary_queue for auxiliary_queue in [auxiliary.storages for auxiliary in auxiliaries ]))
        auxiliary_storages = remove_duplicate_items(auxiliary_storages)
        states = list(
            util.flatten_object([resource.states for resource in self.resources])
        )
        states = remove_duplicate_items(states)

        time_models = (
            [process_instance.time_model for process_instance in processes if not isinstance(process_instance, process.RequiredCapabilityProcess)]
            + [state.time_model for state in states]
            + [source.time_model for source in self.sources]
        )
        time_models += [
            s.repair_time_model
            for s in states
            if isinstance(s, state.BreakDownState)
            or isinstance(s, state.ProcessBreakdownState)
        ]
        time_models += [
            s.battery_time_model
            for s in states
            if isinstance(s, state.ChargingState)
        ]
        time_models = remove_duplicate_items(time_models)

        time_model_data = [time_model.to_model() for time_model in time_models]
        process_data = [process.to_model() for process in processes]
        state_data = [state.to_model() for state in states]
        nodes_data = [node.to_model() for node in nodes]
        product_data = [product.to_model() for product in products]
        resource_data = [resource.to_model() for resource in self.resources]
        source_data = [source.to_model() for source in self.sources]
        sink_data = [sink.to_model() for sink in self.sinks]
        auxiliary_data = [auxiliary.to_model() for auxiliary in auxiliaries]
        auxiliary_storage_data = [storage.to_model() for storage in auxiliary_storages]

        queue_data = list(
            util.flatten_object(
                [s._output_queues for s in self.sources]
                + [
                    r._input_queues
                    for r in self.resources
                    if isinstance(r, resources.ProductionResource)
                ]
                + [
                    r._output_queues
                    for r in self.resources
                    if isinstance(r, resources.ProductionResource)
                ]
                + [s._input_queues for s in self.sinks]
            )
        )
        return json_adapter.JsonProductionSystemAdapter(
            time_model_data=time_model_data,
            process_data=process_data,
            state_data=state_data,
            node_data=nodes_data,
            product_data=product_data,
            nodes_data=nodes_data,
            resource_data=resource_data,
            source_data=source_data,
            sink_data=sink_data,
            queue_data=queue_data + auxiliary_storage_data,
            auxiliary_data = auxiliary_data,
        )

    def run(self, time_range: float = 2880, seed: int = 0):
        """
        Runs the simulation of the production system.
        
        Args:
            time_range (float, optional): The time range of the simulation. Defaults to 2880.
            seed (int, optional): The seed of the simulation. Defaults to 0.
        """
        self._runner = runner.Runner(adapter=self.to_model())
        self._runner.adapter.seed = seed
        self._runner.initialize_simulation()
        self._runner.run(time_range)

    def validate(self):
        """
        Validates the production system. Checks if the production system is valid.

        Raises:
            ValueError: If the production system is not valid.
        """
        adapter = self.to_model()
        adapter.validate_configuration()

    @property
    def runner(self):
        if not self._runner:
            raise ValueError(
                "Runner has not been initialized. Please run the simulation first with the run function."
            )
        return self._runner

    @property
    def post_processor(self):
        if not self._runner:
            raise ValueError(
                "Runner has not been initialized. Please run the simulation first with the run function."
            )
        return self._runner.get_post_processor()



from __future__ import annotations

from dataclasses import field
from typing import List, TYPE_CHECKING

from pydantic import ConfigDict, BaseModel, TypeAdapter

from prodsys.simulation import sim
from prodsys.factories import time_model_factory
from prodsys.models import state_data
from prodsys.simulation import state


if TYPE_CHECKING:
    from prodsys.adapters import adapter


class StateFactory:
    """
    Factory class that creates and stores `prodsys.simulation` state objects from `prodsys.models` state objects.

    Args:
        env (sim.Environment): prodsys simulation environment.
        time_model_factory (time_model_factory.TimeModelFactory): Factory that creates time model objects.
    """

    def __init__(self, env: sim.Environment, time_model_factory: time_model_factory.TimeModelFactory):
        self.env = env
        self.time_model_factory = time_model_factory
        self.state_data: List[state_data.STATE_DATA_UNION] = []
        self.states: List[state.STATE_UNION] = []

    def create_states_from_configuration_data(self, configuration_data: dict):
        for cls_name, items in configuration_data.items():
            for values in items.values():
                values.update({"type": cls_name})
                self.state_data.append(
                    TypeAdapter(state_data.STATE_DATA_UNION).validate_python(values)
                )
                self.add_state(self.state_data[-1])

    def add_state(self, state_data: state_data.STATE_DATA_UNION):
        values = {}
        values.update({"state_data": state_data})
        time_model = self.time_model_factory.get_time_model(
            values["state_data"].time_model_id
        )
        values.update({"time_model": time_model, "env": self.env})
        if "repair_time_model_id" in state_data.model_dump():
            repair_time_model = self.time_model_factory.get_time_model(
                state_data.repair_time_model_id
            )
            values.update({"repair_time_model": repair_time_model})
        if "battery_time_model_id" in state_data.model_dump():
            battery_time_model = self.time_model_factory.get_time_model(
                state_data.battery_time_model_id
            )
            values.update({"battery_time_model": battery_time_model})
        # FIXME: resolve bug when importing simulation types#
        self.states.append(TypeAdapter(state.STATE_UNION).validate_python(values))

    def create_states(self, adapter: adapter.ProductionSystemAdapter):
        """
        Creates state objects based on the given adapter.

        Args:
            adapter (adapter.ProductionSystemAdapter): Adapter that contains the state data.
        """
        for state_data in adapter.state_data:
            self.add_state(state_data)

    def get_states(self, IDs: List[str]) -> List[state.STATE_UNION]:
        """
        Returns a list of state objects with the given IDs.

        Args:
            IDs (List[str]): List of IDs that is used to sort the state objects.

        Returns:
            List[state.STATE_UNION]: List of state objects with the given IDs.
        """
        return [st for st in self.states if st.state_data.ID in IDs]



from copy import deepcopy
import logging
from typing import Callable, List
from prodsys import adapters
from prodsys.adapters.adapter import add_default_queues_to_resources, get_possible_production_processes_IDs, get_possible_transport_processes_IDs, remove_queues_from_resources
from prodsys.models import resource_data, scenario_data
from prodsys.optimization.optimization import check_valid_configuration
from prodsys.optimization.util import add_setup_states_to_machine, adjust_process_capacities, clean_out_breakdown_states_of_resources, get_grouped_processes_of_machine, get_required_auxiliaries


import random
from uuid import uuid1

from prodsys.util.util import flatten


def crossover(ind1, ind2):
    ind1[0].ID = str(uuid1())
    ind2[0].ID = str(uuid1())

    crossover_type = random.choice(["machine", "partial_machine", "transport_resource"])
    adapter1: adapters.ProductionSystemAdapter = ind1[0]
    adapter2: adapters.ProductionSystemAdapter = ind2[0]
    machines_1 = adapters.get_machines(adapter1)
    machines_2 = adapters.get_machines(adapter2)
    remove_queues_from_resources(machines_1 + machines_2)
    transport_resources_1 = adapters.get_transport_resources(adapter1)
    transport_resources_2 = adapters.get_transport_resources(adapter2)
    if "machine" in crossover_type:
        adapter1.resource_data = transport_resources_1
        adapter2.resource_data = transport_resources_2
        if crossover_type == "partial_machine":
            min_length = min(len(machines_1),len(machines_2))
            machines_1 = machines_1[:min_length] + machines_2[min_length:]
            machines_2 = machines_2[:min_length] + machines_1[min_length:]
        adapter1.resource_data += machines_2
        adapter2.resource_data += machines_1

    if crossover_type == "transport_resource":
        adapter1.resource_data = machines_1 + transport_resources_2
        adapter2.resource_data = machines_2 + transport_resources_1

    add_default_queues_to_resources(adapter1)
    add_default_queues_to_resources(adapter2)
    clean_out_breakdown_states_of_resources(adapter1)
    clean_out_breakdown_states_of_resources(adapter2)
    adjust_process_capacities(adapter1)
    adjust_process_capacities(adapter2)

    return ind1, ind2


def add_machine(adapter_object: adapters.ProductionSystemAdapter) -> bool:
    """
    Function that adds a random machine to the production system.

    Args:
        adapter_object (adapters.ProductionSystemAdapter): Production system configuration with specified scenario data.

    Returns:
        bool: True if a machine was added, False otherwise (if adding is not possible due to constraint violations).
    """
    num_process_modules = (
        random.choice(
            range(
                adapter_object.scenario_data.constraints.max_num_processes_per_machine
            )
        )
        + 1
    )
    possible_processes = get_possible_production_processes_IDs(adapter_object)
    if num_process_modules > len(possible_processes):
        num_process_modules = len(possible_processes)
    process_module_list = random.sample(possible_processes, num_process_modules)
    process_module_list = list(flatten(process_module_list))

    control_policy = random.choice(
        adapter_object.scenario_data.options.machine_controllers
    )
    possible_positions = deepcopy(adapter_object.scenario_data.options.positions)
    for resource in adapters.get_machines(adapter_object):
        if resource.location in possible_positions:
            possible_positions.remove(resource.location)
    if not possible_positions:
        return False
    location = random.choice(possible_positions)
    machine_ids = [
        resource.ID
        for resource in adapter_object.resource_data
        if isinstance(resource, resource_data.ProductionResourceData)
    ]
    machine_id = str(uuid1())
    adapter_object.resource_data.append(
        resource_data.ProductionResourceData(
            ID=machine_id,
            description="",
            capacity=1,
            location=location,
            controller=resource_data.ControllerEnum.PipelineController,
            control_policy=control_policy,
            process_ids=process_module_list,
        )
    )
    add_default_queues_to_resources(adapter_object)
    add_setup_states_to_machine(adapter_object, machine_id)
    return True


def add_transport_resource(adapter_object: adapters.ProductionSystemAdapter) -> bool:
    """
    Function that adds a random transport resource to the production system.

    Args:
        adapter_object (adapters.ProductionSystemAdapter): Production system configuration with specified scenario data.

    Returns:
        bool: True if a transport resource was added, False otherwise (if adding is not possible due to constraint violations).
    """
    control_policy = random.choice(
        adapter_object.scenario_data.options.transport_controllers
    )

    transport_resource_ids = [
        resource.ID
        for resource in adapter_object.resource_data
        if isinstance(resource, resource_data.TransportResourceData)
    ]
    transport_resource_id = str(uuid1())
    possible_processes = get_possible_transport_processes_IDs(adapter_object)
    transport_process = random.choice(possible_processes)
    while transport_resource_id in transport_resource_ids:
        transport_resource_id = str(uuid1())
    adapter_object.resource_data.append(
        resource_data.TransportResourceData(
            ID=transport_resource_id,
            description="",
            capacity=1,
            location=(0.0, 0.0),
            controller="TransportController",
            control_policy=control_policy,
            process_ids=[transport_process],
        )
    )
    return True


def add_process_module(adapter_object: adapters.ProductionSystemAdapter) -> bool:
    """
    Function that adds a random process module to a random machine of the production system.

    Args:
        adapter_object (adapters.ProductionSystemAdapter): Production system configuration with specified scenario data.

    Returns:
        bool: True if a process module was added, False otherwise (if adding is not possible due to constraint violations).
    """
    possible_machines = adapters.get_machines(adapter_object)
    if not possible_machines:
        return False
    possible_processes = get_possible_production_processes_IDs(adapter_object)
    machine = random.choice(possible_machines)
    process_module_to_add = random.sample(possible_processes, k=1)
    process_module_to_add = list(flatten(process_module_to_add))
    for process_id in process_module_to_add:
        if process_id not in machine.process_ids:
            machine.process_ids.append(process_id)
    add_setup_states_to_machine(adapter_object, machine.ID)
    return True


def remove_machine(adapter_object: adapters.ProductionSystemAdapter) -> bool:
    """
    Function that removes a random machine from the production system.

    Args:
        adapter_object (adapters.ProductionSystemAdapter): Production system configuration with specified scenario data.

    Returns:
        bool: True if a machine was removed, False otherwise (if removing is not possible due to constraint violations).
    """
    possible_machines = adapters.get_machines(adapter_object)
    if not possible_machines:
        return False
    machine = random.choice(possible_machines)
    adapter_object.resource_data.remove(machine)
    return True


def remove_transport_resource(adapter_object: adapters.ProductionSystemAdapter) -> bool:
    """
    Function that removes a random transport resource from the production system.

    Args:
        adapter_object (adapters.ProductionSystemAdapter): Production system configuration with specified scenario data.

    Returns:
        bool: True if a transport resource was removed, False otherwise (if removing is not possible due to constraint violations).
    """
    transport_resources = adapters.get_transport_resources(adapter_object)
    if not transport_resources:
        return False
    transport_resource = random.choice(transport_resources)
    adapter_object.resource_data.remove(transport_resource)
    return True


def remove_process_module(adapter_object: adapters.ProductionSystemAdapter) -> bool:
    """
    Function that removes a random process module from a random machine of the production system.

    Args:
        adapter_object (adapters.ProductionSystemAdapter): Production system configuration with specified scenario data.

    Returns:
        bool: True if a process module was removed, False otherwise (if removing is not possible due to constraint violations).
    """
    possible_machines = adapters.get_machines(adapter_object)
    if not possible_machines:
        return False
    machine = random.choice(possible_machines)

    possible_processes = get_possible_production_processes_IDs(adapter_object)
    process_modules = get_grouped_processes_of_machine(machine, possible_processes)
    if not process_modules:
        return False
    process_module_to_delete = random.choice(process_modules)

    for process in process_module_to_delete:
        machine.process_ids.remove(process)
    add_setup_states_to_machine(adapter_object, machine.ID)
    return True


def move_process_module(adapter_object: adapters.ProductionSystemAdapter) -> bool:
    """
    Function that moves a random process module from a random machine to another random machine of the production system.

    Args:
        adapter_object (adapters.ProductionSystemAdapter): Production system configuration with specified scenario data.

    Returns:
        bool: True if a process module was moved, False otherwise (if moving is not possible due to constraint violations).
    """
    possible_machines = adapters.get_machines(adapter_object)
    if not possible_machines or len(possible_machines) < 2:
        return False
    from_machine = random.choice(possible_machines)
    possible_machines.remove(from_machine)
    to_machine = random.choice(possible_machines)

    possible_processes = get_possible_production_processes_IDs(adapter_object)
    grouped_process_module_IDs = get_grouped_processes_of_machine(
        from_machine, possible_processes
    )
    if not grouped_process_module_IDs:
        return False
    process_module_to_move = random.choice(grouped_process_module_IDs)
    for process_module in process_module_to_move:
        from_machine.process_ids.remove(process_module)
        to_machine.process_ids.append(process_module)
    add_setup_states_to_machine(adapter_object, from_machine.ID)
    add_setup_states_to_machine(adapter_object, to_machine.ID)
    return True


def move_machine(adapter_object: adapters.ProductionSystemAdapter) -> bool:
    """
    Function that moves a random machine to a random position of the production system.

    Args:
        adapter_object (adapters.ProductionSystemAdapter): Production system configuration with specified scenario data.

    Returns:
        bool: True if a machine was moved, False otherwise (if moving is not possible due to constraint violations).
    """
    possible_machines = adapters.get_machines(adapter_object)
    if not possible_machines:
        return False
    moved_machine = random.choice(possible_machines)
    possible_positions = deepcopy(adapter_object.scenario_data.options.positions)
    for machine in possible_machines:
        if machine.location in possible_positions:
            possible_positions.remove(machine.location)
    if not possible_positions:
        return False
    moved_machine.location = random.choice(possible_positions)
    return True


def add_auxiliary(adapter_object: adapters.ProductionSystemAdapter) -> bool:
    """
    Function that adds a random auxiliary component to the production system.

    Args:
        adapter_object (adapters.ProductionSystemAdapter): Production system configuration with specified scenario data.

    Returns:
        bool: True if an auxiliary component was added, False otherwise (if adding is not possible due to constraint violations).
    """
    required_auxiliaries = get_required_auxiliaries(adapter_object)
    auxiliary = random.choice(required_auxiliaries)
    storage_index = random.choice(range(len(auxiliary.storages)))
    queue = [queue for queue in adapter_object.queue_data if queue.ID == auxiliary.storages[storage_index]][0]
    if queue.capacity == auxiliary.quantity_in_storages[storage_index]:
        return False
    auxiliary.quantity_in_storages[storage_index] += 1
    return True

def remove_auxiliary(adapter_object: adapters.ProductionSystemAdapter) -> bool:
    """
    Function that removes a random auxiliary component from the production system.

    Args:
        adapter_object (adapters.ProductionSystemAdapter): Production system configuration with specified scenario data.

    Returns:
        bool: True if an auxiliary component was removed, False otherwise (if removing is not possible due to constraint violations).
    """
    required_auxiliaries = get_required_auxiliaries(adapter_object)
    auxiliary = random.choice(required_auxiliaries)
    storage_index = random.choice(range(len(auxiliary.storages)))
    if auxiliary.quantity_in_storages[storage_index] == 0:
        return False
    auxiliary.quantity_in_storages[storage_index] -= 1
    return True


def change_control_policy(adapter_object: adapters.ProductionSystemAdapter) -> bool:
    """
    Function that changes the control policy of a random resource of the production system.

    Args:
        adapter_object (adapters.ProductionSystemAdapter): Production system configuration with specified scenario data.

    Returns:
        bool: True if the control policy was changed, False otherwise (if changing is not possible due to constraint violations).
    """
    if not adapter_object.resource_data:
        return False
    resource = random.choice(adapter_object.resource_data)
    if isinstance(resource, resource_data.ProductionResourceData):
        possible_control_policies = deepcopy(
            adapter_object.scenario_data.options.machine_controllers
        )
    else:
        possible_control_policies = deepcopy(
            adapter_object.scenario_data.options.transport_controllers
        )

    if len(possible_control_policies) < 2:
        return False
    possible_control_policies.remove(resource.control_policy)
    new_control_policy = random.choice(possible_control_policies)
    resource.control_policy = new_control_policy
    return True


def change_routing_policy(adapter_object: adapters.ProductionSystemAdapter) -> None:
    """
    Function that changes the routing policy of a random source of the production system.

    Args:
        adapter_object (adapters.ProductionSystemAdapter): Production system configuration with specified scenario data.
    """
    source = random.choice(adapter_object.source_data)
    possible_routing_policies = deepcopy(
        adapter_object.scenario_data.options.routing_heuristics
    )
    if len(possible_routing_policies) < 2:
        return False
    possible_routing_policies.remove(source.routing_heuristic)
    source.routing_heuristic = random.choice(possible_routing_policies)
    return True


def get_mutation_operations(
    adapter_object: adapters.ProductionSystemAdapter,
) -> List[Callable[[adapters.ProductionSystemAdapter], bool]]:
    mutations_operations = []
    transformations = adapter_object.scenario_data.options.transformations
    if scenario_data.ReconfigurationEnum.PRODUCTION_CAPACITY in transformations:
        mutations_operations.append(add_machine)
        mutations_operations.append(remove_machine)
        mutations_operations.append(move_machine)
        mutations_operations.append(change_control_policy)
        mutations_operations.append(add_process_module)
        mutations_operations.append(remove_process_module)
        mutations_operations.append(move_process_module)
    if scenario_data.ReconfigurationEnum.TRANSPORT_CAPACITY in transformations:
        mutations_operations.append(add_transport_resource)
        mutations_operations.append(remove_transport_resource)
    if scenario_data.ReconfigurationEnum.AUXILIARY_CAPACITY in transformations:
        mutations_operations.append(add_auxiliary)
        mutations_operations.append(remove_auxiliary)
    if scenario_data.ReconfigurationEnum.LAYOUT in transformations:
        mutations_operations.append(move_machine)
    if scenario_data.ReconfigurationEnum.SEQUENCING_LOGIC in transformations:
        mutations_operations.append(change_control_policy)
    if scenario_data.ReconfigurationEnum.ROUTING_LOGIC in transformations:
        mutations_operations.append(change_routing_policy)
    return mutations_operations


def mutation(individual):
    mutation_operation = random.choice(get_mutation_operations(individual[0]))
    adapter_object = individual[0]
    if mutation_operation(adapter_object):
        individual[0].ID = str(uuid1())
    add_default_queues_to_resources(adapter_object)
    clean_out_breakdown_states_of_resources(adapter_object)
    adjust_process_capacities(adapter_object)

    return (individual,)


def arrange_machines(adapter_object: adapters.ProductionSystemAdapter) -> None:
    possible_positions = deepcopy(adapter_object.scenario_data.options.positions)
    for machine in adapters.get_machines(adapter_object):
        machine.location = random.choice(possible_positions)
        possible_positions.remove(machine.location)


def get_random_production_capacity(
    adapter_object: adapters.ProductionSystemAdapter,
) -> adapters.ProductionSystemAdapter:
    """
    Function that adds a random number of machines to the production system.

    Args:
        adapter_object (adapters.ProductionSystemAdapter): Production system configuration with specified scenario data.

    Returns:
        adapters.ProductionSystemAdapter: Production system configuration with specified scenario data and added machines.
    """
    num_machines = (
        random.choice(range(adapter_object.scenario_data.constraints.max_num_machines))
        + 1
    )
    adapter_object.resource_data = adapters.get_transport_resources(adapter_object)
    for _ in range(num_machines):
        add_machine(adapter_object)

    return adapter_object


def get_random_transport_capacity(
    adapter_object: adapters.ProductionSystemAdapter,
) -> adapters.ProductionSystemAdapter:
    """
    Function that adds a random number of transport resources to the production system.

    Args:
        adapter_object (adapters.ProductionSystemAdapter): Production system configuration with specified scenario data.

    Returns:
        adapters.ProductionSystemAdapter: Production system configuration with specified scenario data and added transport resources.
    """
    num_transport_resources = (
        random.choice(
            range(adapter_object.scenario_data.constraints.max_num_transport_resources)
        )
        + 1
    )
    adapter_object.resource_data = adapters.get_machines(adapter_object)
    for _ in range(num_transport_resources):
        add_transport_resource(adapter_object)

    return adapter_object


def get_random_auxiliary_capacity(
    adapter_object: adapters.ProductionSystemAdapter,
) -> adapters.ProductionSystemAdapter:
    required_auxiliaries = get_required_auxiliaries(adapter_object)
    available_storage_capacities = {queue.ID: queue.capacity for queue in adapter_object.queue_data}
    for auxiliary in required_auxiliaries:
        for storage_index, storage in enumerate(auxiliary.storages):
            if available_storage_capacities[storage] == 0:
                continue
            random_capacity = random.choice(range(1, available_storage_capacities[storage] + 1))
            if available_storage_capacities[storage] - random_capacity < 0:
                random_capacity = available_storage_capacities[storage]
            auxiliary.quantity_in_storages[storage_index] = random_capacity
            available_storage_capacities[storage] -= random_capacity
    return adapter_object


def get_random_layout(
    adapter_object: adapters.ProductionSystemAdapter,
) -> adapters.ProductionSystemAdapter:
    """
    Function that randomly arranges the machines of the production system.

    Args:
        adapter_object (adapters.ProductionSystemAdapter): Production system configuration with specified scenario data.

    Returns:
        adapters.ProductionSystemAdapter: Production system configuration with specified scenario data and arranged machines.
    """
    possible_positions = deepcopy(adapter_object.scenario_data.options.positions)
    for machine in adapters.get_machines(adapter_object):
        machine.location = random.choice(possible_positions)
        possible_positions.remove(machine.location)
    return adapter_object


def get_random_control_policies(
    adapter_object: adapters.ProductionSystemAdapter,
) -> adapters.ProductionSystemAdapter:
    """
    Function that randomly assigns control policies to the machines and transport resources of the production system.

    Args:
        adapter_object (adapters.ProductionSystemAdapter): Production system configuration with specified scenario data.

    Returns:
        adapters.ProductionSystemAdapter: Production system configuration with specified scenario data and assigned control policies.
    """
    possible_production_control_policies = deepcopy(
        adapter_object.scenario_data.options.machine_controllers
    )
    for machine in adapters.get_machines(adapter_object):
        machine.control_policy = random.choice(possible_production_control_policies)
    possible_transport_control_policies = deepcopy(
        adapter_object.scenario_data.options.transport_controllers
    )
    for transport_resource in adapters.get_transport_resources(adapter_object):
        transport_resource.control_policy = random.choice(
            possible_transport_control_policies
        )
    return adapter_object


def get_random_routing_logic(
    adapter_object: adapters.ProductionSystemAdapter,
) -> adapters.ProductionSystemAdapter:
    """
    Function that randomly assigns routing logics to the sources of the production system.

    Args:
        adapter_object (adapters.ProductionSystemAdapter): Production system configuration with specified scenario data.

    Returns:
        adapters.ProductionSystemAdapter: Production system configuration with specified scenario data and assigned routing logics.
    """
    possible_routing_logics = deepcopy(
        adapter_object.scenario_data.options.routing_heuristics
    )
    for source in adapter_object.source_data:
        source.routing_heuristic = random.choice(possible_routing_logics)
    return adapter_object


def random_configuration(
    baseline: adapters.ProductionSystemAdapter,
) -> adapters.ProductionSystemAdapter:
    """
    Function that creates a random configuration based on a baseline configuration.

    Args:
        baseline (adapters.ProductionSystemAdapter): Baseline configuration.

    Returns:
        adapters.ProductionSystemAdapter: Random configuration based on a baseline configuration.
    """
    transformations = baseline.scenario_data.options.transformations
    invalid_configuration_counter = 0
    while True:
        adapter_object = baseline.model_copy(deep=True)
        adapter_object.ID = str(uuid1())

        if scenario_data.ReconfigurationEnum.PRODUCTION_CAPACITY in transformations:
            get_random_production_capacity(adapter_object)
        if scenario_data.ReconfigurationEnum.TRANSPORT_CAPACITY in transformations:
            get_random_transport_capacity(adapter_object)
        if scenario_data.ReconfigurationEnum.AUXILIARY_CAPACITY in transformations:
            get_random_auxiliary_capacity(adapter_object)
        if (
            scenario_data.ReconfigurationEnum.LAYOUT in transformations
            and scenario_data.ReconfigurationEnum.PRODUCTION_CAPACITY
            not in transformations
        ):
            get_random_layout(adapter_object)
        if scenario_data.ReconfigurationEnum.SEQUENCING_LOGIC in transformations and (
            scenario_data.ReconfigurationEnum.PRODUCTION_CAPACITY not in transformations
            or scenario_data.ReconfigurationEnum.TRANSPORT_CAPACITY
            not in transformations
        ):
            get_random_control_policies(adapter_object)
        if scenario_data.ReconfigurationEnum.ROUTING_LOGIC in transformations:
            get_random_routing_logic(adapter_object)

        add_default_queues_to_resources(adapter_object)
        clean_out_breakdown_states_of_resources(adapter_object)
        adjust_process_capacities(adapter_object)
        if check_valid_configuration(adapter_object, baseline):
            break
        invalid_configuration_counter += 1
        if invalid_configuration_counter % 1000 == 0:
            logging.warning(f"More than {invalid_configuration_counter} invalid configurations were created in a row. Are you sure that the constraints are correct and not too strict?")
    return adapter_object


def random_configuration_with_initial_solution(
    initial_adapters: List[adapters.ProductionSystemAdapter],
) -> adapters.ProductionSystemAdapter:
    """
    Function that creates a random configuration based on an list of initial solutions.

    Args:
        initial_adapters (List[adapters.ProductionSystemAdapter]): List of initial solutions.

    Returns:
        adapters.ProductionSystemAdapter: Random configuration based on an initial solution.
    """
    adapter_object = random.choice(initial_adapters)
    return random_configuration(adapter_object)


from __future__ import annotations

from typing import List, Optional
from uuid import uuid1


from pydantic import Field, conlist
from pydantic.dataclasses import dataclass

from prodsys.express import core, time_model

from prodsys.models import source_data, queue_data
import prodsys

@dataclass
class Source(core.ExpressObject):
    """
    Class that represents a source.

    Args:
        product (product.Product): Product of the source.
        time_model (time_model.TIME_MODEL_UNION): Time model of the source that determines the inter-arrival time of products.
        location (conlist(float, min_length=2, max_length=2)): Location of the source.
        routing_heuristic (source_data.RoutingHeuristic, optional): Routing heuristic of the source. Defaults to source_data.RoutingHeuristic.random.
        ID (str): ID of the source.
    
    Attributes:
        _output_queues (List[queue_data.QueueData]): Output queues of the source.

    Examples:
        Creation of a source with a product, a time model and a location:
        ```py	
        import prodsys.express as psx
        welding_time_model = psx.time_model_data.FunctionTimeModel(
            distribution_function="normal",
            location=20.0,
            scale=5.0,
        )
        welding_process_1 = psx.process.ProductionProcess(
            time_model=welding_time_model,
        )
        welding_process_2 = psx.process.ProductionProcess(
            time_model=welding_time_model,
        )
        transport_time_model = psx.time_model_data.ManhattenDistanceTimeModel(
            speed=10,
            reaction_time= 0.3
        )
        transport_process = psx.process.TransportProcess(
            time_model=transport_time_model,
        )
        product = psx.Product(
            processes=[welding_process_1, welding_process_2],
            transport_process=transport_process
        )
        arrival_time_model = psx.time_model_data.FunctionTimeModel(
            distribution_function="exponential",
            scale=10.0,
        )
        psx.Source(
            product=product,
            time_model=arrival_time_model,
            location=[0.0, 0.0],
        )
        ```
    """
    product: product.Product
    time_model: time_model.TIME_MODEL_UNION
    location: conlist(float, min_length=2, max_length=2) # type: ignore
    routing_heuristic: source_data.RoutingHeuristic = source_data.RoutingHeuristic.random
    ID: Optional[str] = Field(default_factory=lambda: str(uuid1()))

    _output_queues: List[queue_data.QueueData] = Field(default_factory=list, init=False)

    def to_model(self) -> source_data.SourceData:
        """
        Converts the `prodsys.express` object to a data object from `prodsys.models`.

        Returns:
            source_data.SourceData: An instance of the data object.
        """
        source = source_data.SourceData(
            ID=self.ID,
            description="",
            location=self.location,
            product_type=self.product.ID,
            time_model_id=self.time_model.ID,
            routing_heuristic=self.routing_heuristic,
        )
        self._output_queues = [prodsys.adapters.get_default_queue_for_source(source)]
        source.output_queues = [q.ID for q in self._output_queues]
        return source

from prodsys.express import product


"""
This module contains the data structures for the scenario data that is used in optimization to determine the best configuration of a production system. A Scenario constists thereby of:
- `ScenarioConstrainsData`: The constraints of the scenario.
- `ScenarioOptionsData`: The options of the scenario.
- `ScenarioInfoData`: The information of the scenario.
- 'Objectives': The objectives of the scenario.
- `ScenarioData`: The scenario data that contains the constraints, options, information and objectives of the scenario. 
"""
from typing import List, Optional, Dict
from enum import Enum

from pydantic import BaseModel, ConfigDict, field_validator, conlist

from prodsys.models.performance_indicators import KPIEnum
from prodsys.models.resource_data import ResourceControlPolicy, TransportControlPolicy
from prodsys.models.source_data import RoutingHeuristic


class ReconfigurationEnum(str, Enum):
    """
    Enum that represents the different levels of reconfigurations that are possible.

    - ProductionCapacity: Reconfiguration of production capacity (number of machines and their configuration)
    - TransportCapacity: Reconfiguration of transport capacity (number of transport resources and their configuration)
    - AuxiliaryCapacity: Reconfiguration of auxiliary capacity (number of auxiliaries in the system)
    - Layout: Reconfiguration of layout (only position of resources)
    - SequencingLogic: Reconfiguration of sequencing logic (only the control policy of resources)
    - RoutingLogic: Reconfiguration of routing logic (only the routing heuristic of routers)
    """

    PRODUCTION_CAPACITY = "production_capacity"
    TRANSPORT_CAPACITY = "transport_capacity"
    AUXILIARY_CAPACITY = "auxiliary_capacity"
    LAYOUT = "layout"
    SEQUENCING_LOGIC = "sequencing_logic"
    ROUTING_LOGIC = "routing_logic"


class ScenarioConstrainsData(BaseModel):
    """
    Class that represents the constraints of a scenario. The maximum limitations aim to limit the complexity of the
    scenario, thus optimization is possible in a reasonable time. However, by setting only a few constraints, the
    complexity of the scenario can be increased. E.g. you set the maximum reconfiguration cost to the targeted value and
    maximum numbers very high. Thereby, the optimization problem is only constrained by reconfiguration cost and not other
    factors.

    Args:
        max_reconfiguration_cost (float): Maximum reconfiguration cost that can be spend for new production capacity in the scenario.
        max_num_machines (int): Maximum number of machines that can be used in the scenario.
        max_num_processes_per_machine (int): Maximum number of processes that can be assigned to a machine in the scenario.
        max_num_transport_resources (int): Maximum number of transport resources that can be used in the scenario.
        target_product_count (Optional[Dict[str, int]], optional): Target product count for the scenario. Defaults to None. Mapping of product type to target count in the considered time range of the scenario.
    """

    max_reconfiguration_cost: float
    max_num_machines: int
    max_num_processes_per_machine: int
    max_num_transport_resources: int
    target_product_count: Optional[Dict[str, int]]

    model_config=ConfigDict(json_schema_extra={
        "examples": [
            {
                "max_reconfiguration_cost": 120000,
                "max_num_machines": 10,
                "max_num_processes_per_machine": 2,
                "max_num_transport_resources": 2,
                "target_product_count": {"Product_1": 120, "Product_2": 200},
            }
        ]
    })

class ScenarioOptionsData(BaseModel):
    """
    Class that represents the options of a scenario. The options are used to define the deegrees of freedom in the
    optimization scenario, i.e. the different possibilities to adjust the configuration to find a solution with higher
    performance. Options consider possible transformations of the configuration, possible logics of controllers and routers
    and possible positions of machines in the layout.

    Args:
        transformations (List[ReconfigurationEnum]): List of possible transformations of the configuration.
        machine_controllers (List[Literal["FIFO", "LIFO", "SPT"]]): List of possible controllers for machines.
        transport_controllers (List[Literal["FIFO", "SPT_transport", "Nearest_origin_and_longest_target_queues_transport", "Nearest_origin_and_shortest_target_input_queues_transport"]]): List of possible controllers for transport resources.
        routing_heuristics (List[Literal["shortest_queue", "random", "FIFO"]]): List of possible routing heuristics for sources.
        positions (List[conlist(float, min_length=2, max_length=2)]): List of possible positions for machines in the layout.

    Raises:
        ValueError: If the positions are not a list of tuples of length 2.
    """

    transformations: List[ReconfigurationEnum]
    machine_controllers: List[ResourceControlPolicy]
    transport_controllers: List[TransportControlPolicy]
    routing_heuristics: List[RoutingHeuristic]
    positions: List[conlist(float, min_length=2, max_length=2)] # type: ignore

    @field_validator("positions")
    def check_positions(cls, v):
        for e in v:
            if len(e) != 2:
                raise ValueError("positions must be a list of tuples of length 2")
        return v

    model_config=ConfigDict(json_schema_extra={
        "examples": [
            {
                "transformations": [
                    "production_capacity",
                    "transport_capacity",
                    "layout",
                    "sequencing_logic",
                    "routing_logic",
                ],
                "machine_controllers": ["FIFO", "LIFO", "SPT"],
                "transport_controllers": ["FIFO", "SPT_transport"],
                "routing_heuristics": ["shortest_queue", "random", "FIFO"],
                "positions": [[10.0, 10.0], [20.0, 20.0]],
            }
        ]
    })


class ScenarioInfoData(BaseModel):
    """
    Class that represents the information of a scenario. The information is used to define some parameters that allow
    evaluation of the scenario.

    Args:
        machine_cost (float): Cost of a machine.
        transport_resource_cost (float): Cost of a transport resource.
        process_module_cost (float): Cost of a process module.
        breakdown_cost (Optional[float], optional): Cost of a breakdown. Defaults to None.
        time_range (Optional[int], optional): Time range of the scenario in minutes to be considered. Defaults to None.
        maximum_breakdown_time (Optional[int], optional): Maximum allowable breakdown time in the scenario in minutes. Defaults to None.
    """

    machine_cost: float
    transport_resource_cost: float
    process_module_cost: float
    auxiliary_cost: Optional[float] = None
    breakdown_cost: Optional[float] = None
    time_range: Optional[int]
    maximum_breakdown_time: Optional[int] = None

    model_config=ConfigDict(json_schema_extra={
        "examples": [
            {
                "machine_cost": 30000,
                "transport_resource_cost": 20000,
                "process_module_cost": 2300,
                "auxiliary_cost": 1000,
                "breakdown_cost": 1000,
                "time_range": 2600,
                "maximum_breakdown_time": 10,
            }
        ]
    })


class Objective(BaseModel):
    name: KPIEnum
    weight: float = 1.0

    model_config=ConfigDict(json_schema_extra={
        "examples": [
            {
                "name": KPIEnum.COST,
                "weight": 0.6,
            },
            {
                "name": KPIEnum.THROUGHPUT,
                "weight": 0.1,
            },
            {
                "name": KPIEnum.WIP,
                "weight": 0.5,
            },
        ]
    })


class ScenarioData(BaseModel):
    """
    Class that represents a scenario and contraints data about constaints, options, information and objectives of the scenario.

    Args:
        constraints (ScenarioConstrainsData): Constraints of the scenario.
        options (ScenarioOptionsData): Options of the scenario.
        info (ScenarioInfoData): Information of the scenario.
        optimize (List[Objectives]): List of KPIs that should be optimized.

    Raises:
        ValueError: If the weights are not specified for all KPIs that should be optimized.
    """

    constraints: ScenarioConstrainsData
    options: ScenarioOptionsData
    info: ScenarioInfoData
    objectives: List[Objective]

    model_config=ConfigDict(use_enum_values=True, json_schema_extra={
        "examples": [
            {
                "summary": "Scenario",
                "value": {
                    "constraints": {
                        "max_reconfiguration_cost": 120000,
                        "max_num_machines": 10,
                        "max_num_processes_per_machine": 2,
                        "max_num_transport_resources": 2,
                        "target_product_count": {"Product_1": 120, "Product_2": 200},
                    },
                    "options": {
                        "transformations": [
                            "production_capacity",
                            "transport_capacity",
                            "layout",
                            "sequencing_logic",
                            "routing_logic",
                        ],
                        "machine_controllers": ["FIFO", "LIFO", "SPT"],
                        "transport_controllers": ["FIFO", "SPT_transport"],
                        "routing_heuristics": ["shortest_queue", "random", "FIFO"],
                        "positions": [[10.0, 10.0], [20.0, 20.0]],
                    },
                    "info": {
                        "machine_cost": 30000,
                        "transport_resource_cost": 20000,
                        "process_module_cost": 2300,
                        "breakdown_cost": 1000,
                        "time_range": 2600,
                        "maximum_breakdown_time": 10,
                    },
                    "objectives": [
                        {
                            "name": KPIEnum.COST,
                            "weight": 0.6,
                        },
                        {
                            "name": KPIEnum.THROUGHPUT,
                            "weight": 0.1,
                        },
                        {
                            "name": KPIEnum.WIP,
                            "weight": 0.5,
                        },
                    ],
                },
            }
        ]
    })



from prodsys.conf.logging_config import set_logging
from prodsys.models import (
    processes_data,
    product_data,
    queue_data,
    auxiliary_data,
    resource_data,
    scenario_data,
    sink_data,
    source_data,
    state_data,
    time_model_data,
)
from prodsys import adapters
from prodsys.util import post_processing  # , optimization_util
from prodsys.util import runner


VERSION = "0.8.5"



"""
Module containts utility functions for the optimization module. These functions can be used to change the production system configuration and evaluate its performance.
"""

from __future__ import annotations

from typing import Dict, List, Union, Tuple, Literal
from enum import Enum
import logging

from prodsys.express.state import ProcessBreakdownState
from prodsys.models.auxiliary_data import AuxiliaryData



logger = logging.getLogger(__name__)
from pydantic import TypeAdapter

from prodsys import adapters
from prodsys.adapters.adapter import get_possible_production_processes_IDs
from prodsys.models import (
    processes_data,
    resource_data,
    state_data,
    performance_indicators,
    time_model_data
)


class BreakdownStateNamingConvention(str, Enum):
    MACHINE_BREAKDOWN_STATE = "BSM"
    TRANSPORT_RESOURCE_BREAKDOWN_STATE = "BST"
    PROCESS_MODULE_BREAKDOWN_STATE = "BSP"


def get_breakdown_state_ids_of_machine_with_processes(
    processes: List[str], adapter_object: adapters.ProductionSystemAdapter
) -> List[str]:
    state_ids = []
    if check_breakdown_state_available(adapter_object, BreakdownStateNamingConvention.MACHINE_BREAKDOWN_STATE.value):
        state_ids.append(BreakdownStateNamingConvention.MACHINE_BREAKDOWN_STATE.value)
    for process_id in processes:
        process_breakdown_state_id = f"{BreakdownStateNamingConvention.PROCESS_MODULE_BREAKDOWN_STATE.value}_{process_id}"
        if check_breakdown_state_available(adapter_object, process_breakdown_state_id):
            state_ids.append(process_breakdown_state_id)
    return state_ids


def get_required_auxiliaries(
    adapter_object: adapters.ProductionSystemAdapter,
) -> List[AuxiliaryData]:
    """
    Function that returns the required auxiliaries for the production system.

    Args:
        adapter_object (adapters.ProductionSystemAdapter): Production system configuration with specified scenario data.

    Returns:
        List[AuxiliaryData]: List of required auxiliaries
    """
    auxiliary_ids = set()
    for product in adapter_object.product_data:
        auxiliary_ids.update(product.auxiliaries)
    if not auxiliary_ids:
        return []
    return [auxiliary for auxiliary in adapter_object.auxiliary_data if auxiliary.ID in auxiliary_ids or auxiliary.auxiliary_type in auxiliary_ids]


def check_breakdown_state_available(adapter_object: adapters.ProductionSystemAdapter, breakdown_state_id: str) -> bool:
    """
    Function that checks if breakdown states are available in the production system.

    Args:
        adapter_object (adapters.ProductionSystemAdapter): Production system configuration with specified scenario data.

    Returns:
        bool: True if breakdown states are available, False otherwise.
    """
    breakdown_state_ids = set([
        state.ID
        for state in adapter_object.state_data
        if isinstance(state, state_data.BreakDownStateData) or isinstance(state, state_data.ProcessBreakDownStateData)
    ])
    if breakdown_state_id not in breakdown_state_ids:
        return False
    return True

def check_process_breakdown_state_available(adapter_object: adapters.ProductionSystemAdapter) -> bool:
    """
    Function that checks if process breakdown states are available in the production system.

    Args:
        adapter_object (adapters.ProductionSystemAdapter): Production system configuration with specified scenario data.

    Returns:
        bool: True if process breakdown states are available, False otherwise.
    """
    process_ids = [process.ID for process in adapter_object.process_data if isinstance(process, processes_data.ProductionProcessData)]
    for process_id in process_ids:
        process_breakdown_state_id = f"{BreakdownStateNamingConvention.PROCESS_MODULE_BREAKDOWN_STATE.value}_{process_id}"
        if not check_breakdown_state_available(adapter_object, process_breakdown_state_id):
            return False
    return True

def check_breakdown_states_available(adapter_object: adapters.ProductionSystemAdapter) -> bool:
    """
    Function that checks if breakdown states are available in the production system.

    Args:
        adapter_object (adapters.ProductionSystemAdapter): Production system configuration with specified scenario data.

    Returns:
        bool: True if breakdown states are available, False otherwise.
    """
    if (
        not check_breakdown_state_available(adapter_object, BreakdownStateNamingConvention.MACHINE_BREAKDOWN_STATE.value)
        or not check_breakdown_state_available(adapter_object, BreakdownStateNamingConvention.TRANSPORT_RESOURCE_BREAKDOWN_STATE.value)
        or not check_process_breakdown_state_available(adapter_object)
    ):
        return False
    return True


def check_heterogenous_time_models(time_models: List[time_model_data.TIME_MODEL_DATA]) -> bool:
    """
    Function that checks if heterogenous time models are present in the list.

    Args:
        time_models (List[time_model_data.TIME_MODEL_DATA]): List of time models.

    Returns:
        bool: True if heterogenous time models are available, False otherwise.
    """
    # TODO: update this function to use new time models and not the deprecated ones...
    if all(isinstance(time_model, time_model_data.FunctionTimeModelData) for time_model in time_models):
        parameters = []
        for time_model in time_models:
            parameters.append((round(time_model.location), round(time_model.scale), time_model.distribution_function))
        if len(set(parameters)) == 1:
            return True
    elif all(isinstance(time_model, time_model_data.SequentialTimeModelData) for time_model in time_models):
        sequences = []
        for time_model in time_models:
            sequences.append(tuple(time_model.sequence))
        if len(set(sequences)) == 1:
            return True
    elif all(isinstance(time_model, time_model_data.ManhattanDistanceTimeModelData) for time_model in time_models):
        parameters = []
        for time_model in time_models:
            parameters.append((time_model.speed, time_model.reaction_time))
        if len(set(parameters)) == 1:
            return True
    return False


def check_states_for_heterogenous_time_models(states: List[Union[state_data.BreakDownStateData, state_data.ProcessBreakDownStateData]], adapter_object: adapters.ProductionSystemAdapter) -> bool:
    """
    Function that checks if the states have heterogenous time models.

    Args:
        states (List[Union[state_data.BreakDownStateData, state_data.ProcessBreakDownStateData]]): List of states.

    Returns:
        bool: True if the states are compatible, False otherwise.
    """
    all_time_models = adapter_object.time_model_data
    breakdown_time_models = []
    repair_time_models = []
    for state in states:
        breakdown_time_models.append([time_model for time_model in all_time_models if time_model.ID == state.time_model_id].pop())
        repair_time_models.append([time_model for time_model in all_time_models if time_model.ID == state.repair_time_model_id].pop())
    return (check_heterogenous_time_models(breakdown_time_models) and check_heterogenous_time_models(repair_time_models))


def create_default_breakdown_states(adapter_object: adapters.ProductionSystemAdapter):
    logger.info(f"Trying to create default breakdown states.")
    breakdown_states = [
        state
        for state in adapter_object.state_data
        if isinstance(state, state_data.BreakDownStateData)
    ]
    process_breakdown_states = [state for state in adapter_object.state_data if isinstance(state, state_data.ProcessBreakDownStateData)]
    machines = adapters.get_machines(adapter_object)
    transport_resources = adapters.get_transport_resources(adapter_object)
    machine_breakdown_states = [state for state in breakdown_states if any(state.ID in machine.state_ids for machine in machines)]
    transport_resource_breakdown_states = [state for state in breakdown_states if any(state.ID in transport_resource.state_ids for transport_resource in transport_resources)]
    process_breakdown_states = [state for state in process_breakdown_states if any(state.ID in machine.state_ids and state.process_id in machine.process_ids for machine in machines)]
    if machine_breakdown_states and not check_breakdown_state_available(adapter_object, BreakdownStateNamingConvention.MACHINE_BREAKDOWN_STATE.value):
        if not check_states_for_heterogenous_time_models(machine_breakdown_states, adapter_object):
            raise ValueError(f"The machine breakdown states are not heterogenous and it is not ambiguous which state should be the Breakdownstate. Please check the time models or define a distinct machine breakdown state called {BreakdownStateNamingConvention.MACHINE_BREAKDOWN_STATE.value}.")
        machine_breakdown_state = machine_breakdown_states[0].model_copy(deep=True)
        machine_breakdown_state.ID = BreakdownStateNamingConvention.MACHINE_BREAKDOWN_STATE.value
        adapter_object.state_data.append(machine_breakdown_state)
        logger.info(f"Added default breakdown state for production resources to the production system.")
    if transport_resource_breakdown_states and not check_breakdown_state_available(adapter_object, BreakdownStateNamingConvention.TRANSPORT_RESOURCE_BREAKDOWN_STATE.value):
        if not check_states_for_heterogenous_time_models(transport_resource_breakdown_states, adapter_object):
            raise ValueError(f"The transport resource breakdown states are not heterogenous and it is not ambiguous which state should be the Breakdownstate. Please check the time models or define a distinct transport resource breakdown state called {BreakdownStateNamingConvention.TRANSPORT_RESOURCE_BREAKDOWN_STATE.value}.")
        transport_resource_breakdown_state = transport_resource_breakdown_states[0].model_copy(deep=True)
        transport_resource_breakdown_state.ID = BreakdownStateNamingConvention.TRANSPORT_RESOURCE_BREAKDOWN_STATE.value
        adapter_object.state_data.append(transport_resource_breakdown_state)
        logger.info(f"Added default breakdown state for transport resources to the production system.")
    if process_breakdown_states: 
        process_breakdown_states_by_process_id = {}
        for state in process_breakdown_states:
            process_breakdown_states_by_process_id[state.process_id] = process_breakdown_states_by_process_id.get(state.process_id, []) + [state]
        for process_id, process_breakdown_states_for_process in process_breakdown_states_by_process_id.items():
            if check_breakdown_state_available(adapter_object, f"{BreakdownStateNamingConvention.PROCESS_MODULE_BREAKDOWN_STATE.value}_{process_id}"):
                continue
            if not check_states_for_heterogenous_time_models(process_breakdown_states_for_process, adapter_object):
                raise ValueError(f"The process breakdown states are not heterogenous and it is not ambiguous which state should be the Breakdownstate. Please check the time models or define a distinct process breakdown state called {BreakdownStateNamingConvention.PROCESS_MODULE_BREAKDOWN_STATE}.")
            process_breakdown_state: ProcessBreakdownState = process_breakdown_states_for_process[0].model_copy()
            process_breakdown_state.ID = f"{BreakdownStateNamingConvention.PROCESS_MODULE_BREAKDOWN_STATE.value}_{process_id}"
            adapter_object.state_data.append(process_breakdown_state)
            logger.info(f"Added default breakdown state for process modules to the production system.")

def clean_out_breakdown_states_of_resources(
    adapter_object: adapters.ProductionSystemAdapter,
):
    for resource in adapter_object.resource_data:
        if isinstance(resource, resource_data.ProductionResourceData) and any(
            True
            for state in adapter_object.state_data
            if state.ID == BreakdownStateNamingConvention.MACHINE_BREAKDOWN_STATE.value
            or BreakdownStateNamingConvention.PROCESS_MODULE_BREAKDOWN_STATE.value in state.ID
        ):
            resource.state_ids = get_breakdown_state_ids_of_machine_with_processes(
                resource.process_ids, adapter_object
            )
        elif isinstance(resource, resource_data.TransportResourceData) and any(
            True
            for state in adapter_object.state_data
            if state.ID
            == BreakdownStateNamingConvention.TRANSPORT_RESOURCE_BREAKDOWN_STATE.value
        ):
            resource.state_ids = [
                BreakdownStateNamingConvention.TRANSPORT_RESOURCE_BREAKDOWN_STATE.value
            ]


def get_weights(
    adapter: adapters.ProductionSystemAdapter, direction: Literal["min", "max"]
) -> Tuple[float, ...]:
    """
    Get the weights for the objectives of the optimization from an adapter.

    Args:
        adapter (adapters.ProductionSystemAdapter): Production system configuration with specified scenario data.
        direction (Literal[&quot;min&quot;, &quot;max&quot;]): Optimization target direction of the optimizer.

    Returns:
        Tuple[float, ...]: Tuple of weights for the objectives.
    """
    weights = []
    for objective in adapter.scenario_data.objectives:
        kpi: performance_indicators.KPI_UNION = TypeAdapter(performance_indicators.KPI_UNION).validate_python({"name": objective.name})
        if kpi.target != direction:
            weights.append(objective.weight * -1)
        else:
            weights.append(objective.weight)
    return tuple(weights)


def add_setup_states_to_machine(
    adapter_object: adapters.ProductionSystemAdapter, machine_id: str
):
    machine = next(
        resource
        for resource in adapter_object.resource_data
        if resource.ID == machine_id
    )
    no_setup_state_ids = set(
        [
            state.ID
            for state in adapter_object.state_data
            if not isinstance(state, state_data.SetupStateData)
        ]
    )
    machine.state_ids = [
        state for state in machine.state_ids if state in no_setup_state_ids
    ]
    for state in adapter_object.state_data:
        if (
            not isinstance(state, state_data.SetupStateData)
            or state in machine.state_ids
        ):
            continue
        if (
            state.origin_setup in machine.process_ids
            or state.target_setup in machine.process_ids
        ):
            machine.state_ids.append(state.ID)


def get_grouped_processes_of_machine(
    machine: resource_data.ProductionResourceData,
    possible_processes: List[Union[str, Tuple[str, ...]]],
) -> List[Tuple[str]]:
    grouped_processes = []
    for group in possible_processes:
        if isinstance(group, str):
            group = tuple([group])
        for process in machine.process_ids:
            if process in group:
                grouped_processes.append(group)
                break
    return grouped_processes


def get_num_of_process_modules(
    adapter_object: adapters.ProductionSystemAdapter,
) -> Dict[Tuple[str], int]:
    possible_processes = get_possible_production_processes_IDs(adapter_object)
    num_of_process_modules = {}
    for process in possible_processes:
        if isinstance(process, str):
            process = tuple([process])
        num_of_process_modules[process] = 0
    for machine in adapters.get_machines(adapter_object):
        machine_processes = get_grouped_processes_of_machine(
            machine, possible_processes
        )
        for process in machine_processes:
            num_of_process_modules[process] += 1
    return num_of_process_modules


def adjust_process_capacities(
    adapter_object: adapters.ProductionSystemAdapter,
) -> adapters.ProductionSystemAdapter:
    """
    Function that adjusts the process capacities of the production system.

    Args:
        adapter_object (adapters.ProductionSystemAdapter): Production system configuration with specified scenario data.

    Returns:
        adapters.ProductionSystemAdapter: Production system configuration with adjusted process capacities.
    """
    for resource in adapter_object.resource_data:
        resource.process_capacities = [resource.capacity] * len(resource.process_ids)


def document_individual(
    solution_dict: Dict[str, Union[list, str]],
    save_folder: str,
    individual,
):
    adapter_object: adapters.ProductionSystemAdapter = individual[0]
    current_generation = solution_dict["current_generation"]

    if adapter_object.hash() not in solution_dict["hashes"]:
        solution_dict["hashes"][adapter_object.hash()] = {
            "generation": current_generation,
            "ID": adapter_object.ID,
        }
    adapters.JsonProductionSystemAdapter.model_validate(adapter_object).write_data(
        f"{save_folder}/generation_{current_generation}_{adapter_object.ID}.json"
    )









from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, List, Any, TYPE_CHECKING

from pydantic import BaseModel, ConfigDict, Field, TypeAdapter

from prodsys.simulation import sim, sink
from prodsys.models import sink_data
if TYPE_CHECKING:
    from prodsys.factories import product_factory, queue_factory
    from prodsys.adapters import adapter


class SinkFactory(BaseModel):
    """
    Factory class that creates and stores `prodsys.simulation` sink objects from `prodsys.models` sink objects.

    Args:
        env (sim.Environment): prodsys simulation environment.
        product_factory (product_factory.ProductFactory): Factory that creates product objects.
        queue_factory (queue_factory.QueueFactory): Factory that creates queue objects.
    """
    env: sim.Environment
    product_factory: product_factory.ProductFactory
    queue_factory: queue_factory.QueueFactory

    sinks: List[sink.Sink] = Field(default_factory=list, init=False)

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def create_sinks(self, adapter: adapter.ProductionSystemAdapter):
        """
        Creates sink objects based on the given adapter.

        Args:
            adapter (adapter.ProductionSystemAdapter): Adapter that contains the sink data.
        """
        for data in adapter.sink_data:
            self.add_sink(data)

    def add_sink(self, sink_data: sink_data.SinkData):
        values = {
            "env": self.env,
            "data": sink_data,
            "product_factory": self.product_factory,
        }
        sink_object = TypeAdapter(sink.Sink).validate_python(values)
        self.add_queues_to_sink(sink_object)
        self.sinks.append(sink_object)

    def add_queues_to_sink(self, _sink: sink.Sink):
        input_queues = self.queue_factory.get_queues(_sink.data.input_queues)
        _sink.add_input_queues(input_queues)

    def get_sink(self, ID: str) -> sink.Sink:
        """
        Method returns a sink object with the given ID.
        
        Args:
            ID (str): ID of the sink object.
        Returns:
            sink.Sink: Sink object with the given ID.
        """
        return [s for s in self.sinks if s.data.ID == ID].pop()

    def get_sinks(self, IDs: List[str]) -> List[sink.Sink]:
        """
        Method returns a list of sink objects with the given IDs.

        Args:
            IDs (List[str]): List of IDs that is used to sort the sink objects.

        Returns:
            List[sink.Sink]: List of sink objects with the given IDs.
        """
        return [s for s in self.sinks if s.data.ID in IDs]

    def get_sinks_with_product_type(self, __product_type: str) -> List[sink.Sink]:
        """
        Method returns a list of sink objects with the given product type.

        Args:
            __product_type (str): Product type that is used to sort the sink objects.

        Returns:
            List[sink.Sink]: List of sink objects with the given product type.
        """
        return [s for s in self.sinks if __product_type == s.data.product_type]

from prodsys.factories import product_factory, queue_factory   


"""
The `prodsys.factories` package contains the factory classes to create `prodsys.simulation` objects from `prodsys.models` objects in order to initialize a simulation run.

The following modules are available:

- `prodsys.factories.product_factory`: Contains the factory class to create products.
- `prodsys.factories.process_factory`: Contains the factory class to create processes.
- `prodsys.factories.queue_factory`: Contains the factory class to create queues.
- `prodsys.factories.resource_factory`: Contains the factory class to create resources.
- `prodsys.factories.sink_factory`: Contains the factory class to create sinks.
- `prodsys.factories.source_factory`: Contains the factory class to create sources.
- `prodsys.factories.state_factory`: Contains the factory class to create states.
- `prodsys.factories.time_model_factory`: Contains the factory class to create time models.

"""


from __future__ import annotations

from typing import TYPE_CHECKING, List, Dict

from pydantic import BaseModel, ConfigDict, Field


from prodsys.models.product_data import ProductData
from prodsys.simulation import proces_models
from prodsys.simulation import process

if TYPE_CHECKING:
    from prodsys.factories import process_factory
    from prodsys.simulation import sim, router, product



class ProductFactory:
    """
    Factory class that creates and stores `prodsys.simulation` product objects from `prodsys.models` product objects.

    Args:
        env (sim.Environment): prodsys simulation environment.
        process_factory (process_factory.ProcessFactory): Factory that creates process objects.
    """

    def __init__(
        self, env: sim.Environment, process_factory: process_factory.ProcessFactory
    ):
        self.env = env
        self.process_factory = process_factory
        self.products = []
        self.finished_products = []
        self.event_logger = False
        self.product_counter = 0

    def create_product(
        self, product_data: ProductData, router: router.Router
    ) -> product.Product:
        """
        Creates a product object based on the given product data and router.

        Args:
            product_data (ProductData): Product data that is used to create the product object.
            router (router.Router): Router that is used to route the product object.

        Raises:
            ValueError: If the transport process is not found.

        Returns:
            product.Product: Created product object.
        """
        product_data = product_data.model_copy()
        product_data.ID = (
            str(product_data.product_type) + "_" + str(self.product_counter)
        )
        process_model = self.create_process_model(product_data)
        transport_processes = self.process_factory.get_process(
            product_data.transport_process
        )
        if not transport_processes or isinstance(
            transport_processes, process.ProductionProcess
        ):
            raise ValueError("Transport process not found.")
        
        product_object = product.Product(
            env=self.env,
            product_data=product_data,
            product_router=router,
            process_model=process_model,
            transport_process=transport_processes,
        )
        if self.event_logger:
            self.event_logger.observe_terminal_product_states(product_object)

        self.product_counter += 1
        self.products.append(product_object)
        return product_object

    def get_precendece_graph_from_id_adjacency_matrix(
        self, id_adjacency_matrix: Dict[str, List[str]]
    ) -> proces_models.PrecedenceGraphProcessModel:
        precedence_graph = proces_models.PrecedenceGraphProcessModel()
        id_predecessor_adjacency_matrix = (
            proces_models.get_predecessors_adjacency_matrix(id_adjacency_matrix)
        )
        for key in id_adjacency_matrix.keys():
            sucessor_ids = id_adjacency_matrix[key]
            predecessor_ids = id_predecessor_adjacency_matrix[key]
            process = self.process_factory.get_process(key)
            successors = [
                self.process_factory.get_process(successor_id)
                for successor_id in sucessor_ids
            ]
            predecessors = [
                self.process_factory.get_process(predecessor_id)
                for predecessor_id in predecessor_ids
            ]
            precedence_graph.add_node(process, successors, predecessors)
        return precedence_graph

    def create_process_model(
        self, product_data: ProductData
    ) -> proces_models.ProcessModel:
        """
        Creates a process model based on the given product data.

        Args:
            product_data (ProductData): Product data that is used to create the process model.

        Raises:
            ValueError: If the process model is not recognized.

        Returns:
            proces_models.ProcessModel: Created process model.
        """
        if isinstance(product_data.processes, list) and isinstance(
            product_data.processes[0], str
        ):
            process_list = self.process_factory.get_processes_in_order(
                product_data.processes
            )
            return proces_models.ListProcessModel(process_list=process_list)
        elif isinstance(product_data.processes, dict):
            return self.get_precendece_graph_from_id_adjacency_matrix(
                product_data.processes
            )
        elif isinstance(product_data.processes, list) and isinstance(
            product_data.processes[0], list
        ):
            id_adjacency_matrix = proces_models.get_adjacency_matrix_from_edges(
                product_data.processes
            )
            return self.get_precendece_graph_from_id_adjacency_matrix(
                id_adjacency_matrix
            )
        else:
            raise ValueError("Process model not recognized.")

    def get_product(self, ID: str) -> product.Product:
        """
        Returns the product object with the given ID.

        Args:
            ID (str): ID of the product object.

        Returns:
            product.Product: Product object with the given ID.
        """
        return [m for m in self.products if m.product_data.ID == ID].pop()

    def remove_product(self, product: product.Product):
        """
        Removes the given product object from the product factory list of current product objects.

        Args:
            product (product.Product): Product object that is removed.
        """
        self.products = [
            m for m in self.products if m.product_data.ID != product.product_data.ID
        ]

    def register_finished_product(self, product: product.Product):
        """
        Registers the given product object as a finished product object.

        Args:
            product (product.Product): Product object that is registered as a finished product object.
        """
        self.finished_products.append(product)
        self.remove_product(product)

from prodsys.simulation import product



from enum import Enum
from typing import Literal, List, Optional, Union, Tuple

from pydantic import BaseModel, ConfigDict, field_validator


class KPIEnum(str, Enum):
    """
    Enum that represents the different kind of KPIs.
    """

    OUTPUT = "output"
    THROUGHPUT = "throughput"
    COST = "cost"
    WIP = "WIP"
    AUXILIARY_WIP = "auxiliary_WIP"

    TRHOUGHPUT_TIME = "throughput_time"
    PROCESSING_TIME = "processing_time"

    PRODUCTIVE_TIME = "productive_time"
    STANDBY_TIME = "standby_time"
    SETUP_TIME = "setup_time"
    CHARGING_TIME = "charging_time"
    UNSCHEDULED_DOWNTIME = "unscheduled_downtime"

    DYNAMIC_WIP = "dynamic_WIP"
    DYNAMIC_THROUGHPUT_TIME = "dynamic_throughput_time"


class KPILevelEnum(str, Enum):
    """
    Enum that represents the different kind of KPI levels.
    """

    SYSTEM = "system"
    RESOURCE = "resource"
    ALL_PRODUCTS = "all_products"
    PRODUCT_TYPE = "product_type"
    PRODUCT = "product"
    PROCESS = "process"


class KPI(BaseModel):
    """
    Class that represents a KPI. Not intended for usage but only inheritance.

    Args:
        name (KPIEnum): Name of the KPI.
        target (Literal["min", "max"]): Favourable target of the KPI.
        weight (Optional[float], optional): Weight of the KPI. Defaults to 1.
        value (Optional[float], optional): Value of the KPI. Defaults to None.
        context (Tuple[KPILevelEnum, ...], optional): Context of the KPI. Defaults to None.
        resource (Optional[str], optional): Resource of the KPI. Defaults to None.
        product_type (Optional[str], optional): Product type of the KPI. Defaults to None.    Returns:
    """

    name: KPIEnum
    target: Literal["min", "max"]
    weight: Optional[float] = 1
    value: Optional[float] = None
    context: Optional[Tuple[KPILevelEnum, ...]] = None
    resource: Optional[str] = None
    product_type: Optional[str] = None

    @field_validator("context", mode="before")
    def sort_context(cls, v):
        return tuple(sorted(v))


class DynamicKPI(KPI):
    """
    Class that represents a dynamic KPI. Not intended for usage but only inheritance.
    """

    start_time: float
    end_time: float
    product: Optional[str] = None
    process: Optional[str] = None


class Output(KPI):
    name: Literal[KPIEnum.OUTPUT] = KPIEnum.OUTPUT
    target: Literal["max"] = "max"

    model_config = ConfigDict(
        json_schema_extra={
            "examples": [
                {
                    "name": "output",
                    "target": "max",
                    "weight": 1,
                    "value": 34,
                    "context": ["system", "product_type"],
                    "product_type": "ProductType_1",
                }
            ]
        }
    )


class Throughput(KPI):
    name: Literal[KPIEnum.THROUGHPUT]
    target: Literal["max"] = "max"

    model_config = ConfigDict(
        json_schema_extra={
            "examples": [
                {
                    "name": "throughput",
                    "target": "max",
                    "weight": 1,
                    "value": 4.32,
                    "context": ["system", "product_type"],
                    "product_type": "ProductType_1",
                }
            ]
        }
    )

class Cost(KPI):
    name: Literal[KPIEnum.COST]
    target: Literal["min"] = "min"

    model_config = ConfigDict(
        json_schema_extra={
            "examples": [
                {
                    "name": "cost",
                    "target": "min",
                    "weight": 0.5,
                    "value": 36000,
                    "context": ["system"],
                }
            ]
        }
    )


class WIP(KPI):
    name: Literal[KPIEnum.WIP]
    target: Literal["min"] = "min"

    model_config = ConfigDict(
        json_schema_extra={
            "examples": [
                {
                    "name": "WIP",
                    "target": "min",
                    "weight": 1,
                    "value": 121,
                    "context": ["system", "product_type"],
                    "product_type": "ProductType_1",
                }
            ]
        }
    )

    
class AuxiliaryWIP(KPI):
    name: Literal[KPIEnum.AUXILIARY_WIP]
    target: Literal["min"] = "min"

    model_config = ConfigDict(
        json_schema_extra={
            "examples": [
                {
                    "name": "AUXILIARY_WIP",
                    "target": "min",
                    "weight": 1,
                    "value": 121,
                    "context": ["system", "product_type"],
                    "product_type": "Auxiliary_1",
                }
            ]
        }
    )

class DynamicWIP(DynamicKPI, WIP):
    name: Literal[KPIEnum.DYNAMIC_WIP]
    target: Literal["min"] = "min"

    model_config = ConfigDict(json_schema_extra={
        "examples": [
            {
                "name": "dynamic_WIP",
                "target": "min",
                "weight": 1,
                "value": 121,
                "context": ["system", "product"],
                "product_type": "ProductType_1",
                "start_time": 21.2,
                "end_time": 23.4,
            }
        ]
    })


class ThroughputTime(KPI):
    name: Literal[KPIEnum.TRHOUGHPUT_TIME]
    target: Literal["min"] = "min"

    model_config = ConfigDict(json_schema_extra={
        "examples": [
            {
                "name": "throughput_time",
                "target": "min",
                "weight": 1,
                "value": 221.1,
                "context": ["system", "product_type"],
                "product_type": "ProductType_1",
            }
        ]
    })


class DynamicThroughputTime(DynamicKPI, ThroughputTime):
    name: Literal[KPIEnum.DYNAMIC_THROUGHPUT_TIME]
    target: Literal["min"] = "min"

    model_config = ConfigDict(json_schema_extra={
        "examples": [
            {
                "name": "dynamic_throughput_time",
                "target": "min",
                "weight": 1,
                "value": 221.1,
                "context": ["system", "product"],
                "product_type": "ProductType_1",
                "product": "Product_1_23",
                "start_time": 21.2,
                "end_time": 23.4,
            }
        ]
    })


class ProcessingTime(KPI):
    name: Literal[KPIEnum.PROCESSING_TIME]
    target: Literal["min"] = "min"

    model_config = ConfigDict(json_schema_extra={
        "examples": [
            {
                "name": "processing_time",
                "target": "min",
                "weight": 1,
                "value": 1.2,
                "context": ["resource", "process"],
                "resource": "Resource_1",
                "process": "P1",
            }
        ]
    })


class ProductiveTime(KPI):
    name: Literal[KPIEnum.PRODUCTIVE_TIME]
    target: Literal["max"] = "max"

    model_config = ConfigDict(json_schema_extra={
        "examples": [
            {
                "name": "productive_time",
                "target": "max",
                "weight": 1,
                "value": 0.65,
                "context": ["resource"],
                "resource": "Resource_1",
            }
        ]
    })


class StandbyTime(KPI):
    name: Literal[KPIEnum.STANDBY_TIME]
    target: Literal["min"] = "min"

    model_config = ConfigDict(json_schema_extra={
        "examples": [
            {
                "name": "standby_time",
                "target": "min",
                "weight": 1,
                "value": 0.12,
                "context": ["resource"],
                "resource": "Resource_1",
            }
        ]
    })


class SetupTime(KPI):
    name: Literal[KPIEnum.SETUP_TIME]
    target: Literal["min"] = "min"

    model_config = ConfigDict(json_schema_extra={
        "examples": [
            {
                "name": "setup_time",
                "target": "min",
                "weight": 1,
                "value": 0.08,
                "context": ["resource"],
                "resource": "Resource_1",
            }
        ]
    })


class ChargingTime(KPI):
    name: Literal[KPIEnum.CHARGING_TIME]
    target: Literal["min"] = "min"

    model_config = ConfigDict(json_schema_extra={
        "examples": [
            {
                "name": "charging_time",
                "target": "min",
                "weight": 1,
                "value": 0.1,
                "context": ["resource"],
                "resource": "Resource_1",
            }
        ]
    })


class UnscheduledDowntime(KPI):
    name: Literal[KPIEnum.UNSCHEDULED_DOWNTIME]
    target: Literal["min"] = "min"

    model_config = ConfigDict(json_schema_extra={
        "examples": [
            {
                "name": "unscheduled_downtime",
                "target": "min",
                "weight": 1,
                "value": 0.1,
                "context": ["resource"],
                "resource": "Resource_1",
            }
        ]
    })


KPI_UNION = Union[
    Output,
    Throughput,
    Cost,
    WIP,
    ThroughputTime,
    ProcessingTime,
    ProductiveTime,
    StandbyTime,
    SetupTime,
    ChargingTime,
    UnscheduledDowntime,
    DynamicWIP,
    DynamicThroughputTime,
]



from __future__ import annotations

from collections.abc import Callable
from typing import List, TYPE_CHECKING, Generator, Optional, Union

import logging

import simpy
logger = logging.getLogger(__name__)

import numpy as np

from simpy import events

from prodsys.simulation import resources
from prodsys.simulation import request


if TYPE_CHECKING:
    from prodsys.simulation import resources, product, sink, auxiliary
    from prodsys.factories import resource_factory, sink_factory, auxiliary_factory
    from prodsys.control import routing_control_env
    from prodsys.models import product_data
    from prodsys.simulation.product import Locatable


def get_env_from_requests(requests: List[request.Request]) -> simpy.Environment:
    """
    Returns the environment from a list of requests.

    Args:
        requests (List[request.Request]): The requests.

    Returns:
        simpy.Environment: The environment.
    """
    if not requests:
        raise ValueError("No requests found to retrieve an environment from.")
    if requests[0].product:
        return requests[0].product.env
    else:
        return requests[0].auxiliary.env


def get_item_to_transport(request_for_transport: Union[request.Request, request.AuxiliaryRequest]) -> Union[product.Product, auxiliary.Auxiliary]:
    """
    Returns the item to transport from a request.

    Args:
        request_for_transport (Union[request.Request, request.AuxiliaryRequest]): The request.

    Returns:
        Union[product.Product, auxiliary.Auxiliary]: The item to transport.
    """
    if isinstance(request_for_transport, request.AuxiliaryRequest):
        return request_for_transport.auxiliary
    return  request_for_transport.product


class Router:
    """
    Base class for all routers.

    Args:
        resource_factory (resource_factory.ResourceFactory): The resource factory of the production system.
        sink_factory (sink_factory.SinkFactory): The sink factory of the production system.
        routing_heuristic (Callable[[List[resources.Resource]], resources.Resource]): The routing heuristic to be used, needs to be a callable that takes a list of resources and returns a resource.
    """

    def __init__(
        self,
        resource_factory: resource_factory.ResourceFactory,
        sink_factory: sink_factory.SinkFactory,
        auxiliary_factory: auxiliary_factory.AuxiliaryFactory,
        routing_heuristic: Callable[[List[request.Request]], None],
    ):
        self.resource_factory: resource_factory.ResourceFactory = resource_factory
        self.sink_factory: sink_factory.SinkFactory = sink_factory
        self.auxiliary_factory: auxiliary_factory.AuxiliaryFactory = auxiliary_factory
        self.routing_heuristic: Callable[[List[request.Request]], None] = routing_heuristic
        # TODO: add possibility to specify a production and a transport heuristic separately
    
    def route_product_to_production_resource(self, product: product.Product) -> Generator[Optional[request.Request]]:
        """
        Routes a product to perform the next process by assigning a production resource, that performs the process, to the product 
        and assigning a transport resource to transport the product to the next process.

        Args:
            product (product.Product): The product.

        Returns:
            Generator[Tuple[request.Request, request.TransportResquest]]: A generator that yields when the product is routed.
        """
        potential_production_requests = self.get_possible_production_requests(product)
        if not potential_production_requests:
            raise ValueError(f"No possible production resources found for product {product.product_data.ID} and process {product.next_prodution_process.process_data.ID}.")
        potential_transport_requests: List[request.Request] = []

        potential_transport_requests = self.get_possible_transport_requests(potential_production_requests)
        if not potential_transport_requests:
            raise ValueError(f"No possible transport resources found for product {product.product_data.ID} and process {product.next_prodution_process.process_data.ID} to reach any destinations from resource {product.current_locatable.data.ID}.")
        possible_production_requests = self.get_reachable_production_requests(potential_production_requests, potential_transport_requests)
        
        env = get_env_from_requests(potential_transport_requests)
        routed_production_request = yield env.process(self.get_routed_production_request(possible_production_requests))
        return routed_production_request
    

    def get_input_queue_state_change_events(self, possible_requests: List[request.Request]) -> List[events.Event]:
        """
        Returns the state change events of the input queues of the resources of the requests.

        Args:
            possible_requests (List[request.Request]): The requests.

        Returns:
            List[events.Event]: The state change events of the input queues of the resources of the requests.
        """
        possible_resources: List[resources.ProductionResource] = []
        for request in possible_requests:
            if not request.resource in possible_resources and isinstance(request.resource, resources.ProductionResource):
                possible_resources.append(request.resource)
        input_queue_get_events = []
        for resource in possible_resources:
            for queue in resource.input_queues:
                input_queue_get_events.append(queue.state_change)
        return input_queue_get_events
    

    def get_routed_production_request(self, possible_production_requests: List[request.Request]) -> Generator[Optional[request.Request]]:
        env = get_env_from_requests(possible_production_requests)
        while True:
            production_requests: List[request.Request] = self.get_requests_with_non_blocked_resources(possible_production_requests)
            if production_requests:
                break
            logger.debug({"ID": possible_production_requests[0].product.product_data.ID, "sim_time": env.now, "event": f"Waiting for free resources."})
            yield events.AnyOf(
                env,
                self.get_input_queue_state_change_events(possible_production_requests),
            )
            logger.debug({"ID": possible_production_requests[0].product.product_data.ID, "sim_time": env.now, "event": f"Free resources available."})

        self.routing_heuristic(production_requests)
        if not production_requests:
            return
        routed_production_request = production_requests.pop(0)
        routed_production_request.resource.reserve_input_queues()
        return routed_production_request 

    def route_transport_resource_for_item(self, routed_production_request: Union[request.Request, request.AuxiliaryRequest]) -> Generator[Optional[request.TransportResquest]]:
        """
        Routes a product to perform the next process by assigning a production resource, that performs the process, to the product 
        and assigning a transport resource to transport the product to the next process.

        Args:
            item_to_transport (Union[product.Product, auxiliary.Auxiliary]): The product or auxiliary to transport.
            routed_production_request (Union[request.Request, request.AuxiliaryRequest]): The production request or auxiliary request.

        Returns:
            Generator[Optional[request.TransportResquest]]: A generator that yields when the product is routed.
        """
        item_to_transport = get_item_to_transport(routed_production_request)
        potential_transport_requests: List[request.Request] = self.get_transport_requests_to_target(item_to_transport, routed_production_request.resource, {})
        if not potential_transport_requests:
            raise ValueError(f"No possible transport resources found for product {item_to_transport.product_data.ID} and process {item_to_transport.next_prodution_process.process_data.ID} to reach any destinations from resource {item_to_transport.current_locatable.data.ID}.")
        
        env = get_env_from_requests(potential_transport_requests)
        while True:
            transport_requests: List[request.TransportResquest] = self.get_requests_with_non_blocked_resources(potential_transport_requests)
            if transport_requests:
                break
            logger.debug({"ID": item_to_transport.product_data.ID, "sim_time": env.now, "event": f"Waiting for free resources."})
            yield events.AnyOf(
                env,
                self.get_input_queue_state_change_events(potential_transport_requests),
            )
            logger.debug({"ID": item_to_transport.product_data.ID, "sim_time": env.now, "event": f"Free resources available."})
        if not transport_requests:
            raise ValueError(f"No transport requests found for routing of product {item_to_transport.product_data.ID}. Error in Event handling of routing to resources.")
        self.routing_heuristic(transport_requests)
        yield env.timeout(0)
        if not transport_requests:
            return
        routed_transport_request = transport_requests.pop(0)
        return routed_transport_request
    

    def route_product_to_sink(self, product: product.Product) -> Generator[Optional[request.TransportResquest]]:
        """
        Routes a product to a sink.

        Args:
            product (product.Product): The product.

        Returns:
            Generator[request.TransportResquest]: A generator that yields when the product is routed to the sink.
        """
        sink = self.get_sink(product.product_data.product_type)
        sink_request = request.SinkRequest(product=product, sink=sink)
        potential_transport_requests = self.get_possible_transport_requests([sink_request])
        if not potential_transport_requests:
            raise ValueError(f"No possible transport resources found for product {product.product_data.ID} to reach any sinks from resource {product.current_locatable.data.ID}.")
        env = get_env_from_requests(potential_transport_requests)
        while True:
            transport_requests: List[request.TransportResquest] = self.get_requests_with_non_blocked_resources(potential_transport_requests)
            if transport_requests:
                break
            logger.debug({"ID": product.product_data.ID, "sim_time": env.now, "event": f"Waiting for free resources."})
            yield events.AnyOf(
                env,
                self.get_input_queue_state_change_events(potential_transport_requests),
            )
            logger.debug({"ID": product.product_data.ID, "sim_time": env.now, "event": f"Free resources available."})
        if not transport_requests:
            raise ValueError(f"No transport requests found for routing of product {product.product_data.ID}. Error in Event handling of routing to resources.")
        self.routing_heuristic(transport_requests)
        yield env.timeout(0)
        if not transport_requests:
            return
        routed_transport_request = transport_requests.pop(0)
        return routed_transport_request

    def get_production_request(self, product: product.Product, resource: resources.Resource) -> request.Request:
        """
        Returns a request for the next production process of the product object.

        Returns:
            request.Request: The request for the next production process.
        """
        return request.Request(
            process=product.next_prodution_process,
            product=product,
            resource=resource,
        )
    
    def get_transport_request(self, item_to_transport: Union[product.Product, auxiliary.Auxiliary], transport_resource: resources.TransportResource, target: resources.Resource) -> request.TransportResquest:
        """
        Returns a request for the next transport process of the product object.

        Returns:
            request.Request: The request for the next transport process.
        """
        return request.TransportResquest(
            process=item_to_transport.transport_process,
            product=item_to_transport,
            resource=transport_resource,
            origin=item_to_transport.current_locatable,
            target=target,
        )
    

    def route_auxiliary_to_product(self, product: product.Product) -> Generator[request.AuxiliaryRequest]:
        """
        Routes an auxiliary to a product.

        Args:
            product (product.Product): The product.

        Returns:
            Generator: A generator that yields when the auxiliary is routed to the product.
        """
        auxiliary_request = request.AuxiliaryRequest(product=product, process=product.transport_process)
        yield product.env.process(self.get_auxiliary(auxiliary_request))
        yield product.env.timeout(0)
        return auxiliary_request
    
    def route_auxiliary_to_store(self, auxiliary: auxiliary.Auxiliary) -> Generator[request.TransportResquest]:
        """
        Routes an auxiliary to a store.

        Args:
            auxiliary (auxiliary.Auxiliary): The auxiliary.

        Returns:
            Generator: A generator that yields when the auxiliary is routed to the store.
        """
        auxiliary_request = request.AuxiliaryRequest(auxiliary=auxiliary, process=auxiliary.transport_process, resource=auxiliary.storage, product=None)
        # transport_request = self.get_transport_request(auxiliary, auxiliary.transport_process, auxiliary.storage)
        env = get_env_from_requests([auxiliary_request])
        transport_request: request.TransportResquest = yield env.process(self.route_transport_resource_for_item(auxiliary_request))
        yield env.timeout(0)
        return transport_request


    def get_auxiliary(self, processing_request: request.AuxiliaryRequest) -> Generator:

        possible_auxiliaries = self.get_possible_auxiliaries(processing_request)
        while True:
            free_possible_auxiliaries = self.get_free_auxiliary(possible_auxiliaries)
            # TODO: maybe make heuristic working for selecting auxiliary...
            np.random.shuffle(free_possible_auxiliaries)
            if free_possible_auxiliaries:
                break
            logger.info({"ID": processing_request.product.product_data.ID , "sim_time": processing_request.product.env.now, "event": f"Waiting for free auxiliary."})
            yield events.AnyOf(
                processing_request.product.env,
                [auxiliary.got_free for auxiliary in possible_auxiliaries],
            )
            logger.info({"ID": processing_request.product.product_data.ID, "sim_time": processing_request.product.env.now, "event": f"Free auxiliary available."})
        routed_auxiliary = free_possible_auxiliaries[0]
        routed_auxiliary.reserve()
        routed_auxiliary.got_free = events.Event(processing_request.product.env)
        routed_auxiliary.current_product = processing_request.product
        processing_request.auxiliary = routed_auxiliary
        processing_request.resource = processing_request.product.current_locatable

    def get_free_auxiliary(self, possible_auxiliaries: List[auxiliary.Auxiliary]) -> List[auxiliary.Auxiliary]:

        free_possible_auxiliaries = []
        for auxiliary in possible_auxiliaries:
            if not auxiliary.reserved and auxiliary.current_product is None:
                free_possible_auxiliaries.append(auxiliary)
        return free_possible_auxiliaries
    
    def get_possible_auxiliaries(self, processing_request: request.AuxiliaryRequest)-> List[auxiliary.Auxiliary]:
        possible_auxiliaries = []
        for auxiliary in self.auxiliary_factory.auxiliaries:
            if not auxiliary.data.auxiliary_type in processing_request.product.product_data.auxiliaries:
                continue
            if any(supported_process.matches_request(processing_request) for supported_process in auxiliary.relevant_processes + auxiliary.relevant_transport_processes):
                possible_auxiliaries.append(auxiliary)
        return possible_auxiliaries
    
    def get_possible_production_requests(self, product: product.Product) -> List[request.Request]:
        """
        Returns a list of possible production requests with different resources and processes for the next production process of a product.

        Args:
            product (product.Product): The product to get the request for.

        Returns:
            List[request.Request]: A list of possible production requests for the next production process of the product.
        """
        possible_requests = []
        for resource in self.resource_factory.get_production_resources():
            for process in resource.processes:
                production_request = self.get_production_request(product, resource)
                if process.matches_request(production_request):
                    production_request.set_process(process)
                    possible_requests.append(production_request)
        return possible_requests
    
    def get_reachable_production_requests(self, production_requests: List[request.Request], transport_requests: List[request.TransportResquest]) -> List[request.Request]:
        """
        Returns a list of production requests that are reachable by the transport requests.

        Args:
            production_requests (List[request.Request]): potential production requests
            transport_requests (List[request.Request]): potential transport requests

        Returns:
            List[request.Request]: A list of production requests that are reachable by the transport requests.
        """
        possible_production_resource_ids = set([request.target.data.ID for request in transport_requests])
        return [request for request in production_requests if request.resource.data.ID in possible_production_resource_ids]
    
    def get_possible_transport_requests(self, production_requests: List[request.Request]) -> List[request.TransportResquest]:
        """
        Returns a list of possible transport requests with different resources and processes for the next transport process of a product.

        Args:
            production_request (request.Request): The production request to get the transport request for.

        Returns:
            List[request.TransportResquest]: A list of possible transport requests for the next transport process of the product.
        """
        if any(isinstance(production_request, request.SinkRequest) for production_request in production_requests):
            transport_targets = [request.resource for request in production_requests]
        else:
            transport_target_ids = set([request.resource.data.ID for request in production_requests])
            transport_targets = [resource for resource in self.resource_factory.resources if resource.data.ID in transport_target_ids]
        product = production_requests[0].product

        possible_requests = []
        route_cache = {}

        for transport_target in transport_targets:
            possible_requests += self.get_transport_requests_to_target(product, transport_target, route_cache)
        return possible_requests
    


    def get_transport_requests_to_target(self, item_to_transport: Union[product.Product, auxiliary.Auxiliary], target: Locatable, route_cache: dict) -> List[request.TransportResquest]:
        """
        Returns a list of transport requests with different resources and processes for the next transport process of a product.

        Args:
            transport_target (resources.Resource): The transport target to get the transport requests for.

        Returns:
            List[request.TransportResquest]: A list of transport requests for the transport target.
        """
        transport_requests = []
        for resource in self.resource_factory.get_transport_resources():
            for process in resource.processes:
                transport_request = self.get_transport_request(item_to_transport, resource, target)
                if route_cache.get((target.data.ID, process.process_data.ID)):
                    transport_request.copy_cached_routes(route_cache[(target.data.ID, process.process_data.ID)])
                    transport_request.set_process(process)
                    transport_requests.append(transport_request)
                elif process.matches_request(transport_request):
                    transport_request.set_process(process)
                    transport_requests.append(transport_request)
                    route_cache[(target.data.ID, process.process_data.ID)] = transport_request
        return transport_requests
        

    def get_requests_with_non_blocked_resources(
        self, requests: List[request.Request]
    ) -> List[request.Request]:
        """
        Returns a list of requests with non-blocked resources.

        Args:
            requests (List[request.Request]): A list of requests.

        Returns:
            List[request.Request]: A list of requests with non-blocked resources.
        """
        for request in requests:
            if not isinstance(request.resource, resources.ProductionResource):
                continue
            env = get_env_from_requests([request])
        return [request for request in requests if isinstance(request.resource, resources.TransportResource) or (isinstance(request.resource, resources.ProductionResource) and not any(q.full for q in request.resource.input_queues))]

    def get_sink(self, _product_type: str) -> sink.Sink:
        """
        Returns the sink for a product type.

        Args:
            _product_type (str): The product type.

        Returns:
            sink.Sink: The sink for the product type.
        """
        possible_sinks = self.sink_factory.get_sinks_with_product_type(_product_type)
        chosen_sink = np.random.choice(possible_sinks)
        return chosen_sink  # type: ignore False


def FIFO_routing_heuristic(possible_requests: List[request.Request]):
    """
    Sorts the list by the FIFO principle.

    Args:
        possible_resources (List[resources.Resource]): A list of possible resources.
    """
    pass


def random_routing_heuristic(possible_requests: List[request.Request]):
    """
    Shuffles the list of possible resources.

    Args:
        possible_resources (List[resources.Resource]): A list of possible resources.
    """
    possible_requests.sort(key=lambda x: x.resource.data.ID)
    np.random.shuffle(possible_requests)


def shortest_queue_routing_heuristic(
    possible_requests: List[request.Request],
):
    """
    Sorts the list of possible resources by the length of their input queues and returns the first resource.
    For Transport resources, the next resource is chosen by the resource with the shortest request queue.

    Args:
        possible_resources (List[resources.Resource]): A list of possible resources.
    """
    if any(not isinstance(request.resource, resources.ProductionResource) for request in possible_requests):
        np.random.shuffle(possible_requests)
        possible_requests.sort(key=lambda x: len(x.resource.get_controller().requests))
        return
    np.random.shuffle(possible_requests)
    possible_requests.sort(key=lambda x: sum([len(q.items) for q in x.resource.input_queues]))


def agent_routing_heuristic(
    gym_env: routing_control_env.AbstractRoutingControlEnv,
    possible_requests: List[request.Request],
):
    """
    Sorts the list of possible resources by an reinforcement learning agent

    Args:
        gym_env (gym_env.ProductionRoutingEnv): Environment for the reinforcement learning agent.
        possible_resources (List[resources.Resource]): A list of possible resources.
    """
    if gym_env.interrupt_simulation_event.triggered:
        # Avoid that multiple requests trigger event multiple times -> delay them
        while possible_requests:
            possible_requests.pop()
        return
    gym_env.set_possible_requests(possible_requests)
    gym_env.interrupt_simulation_event.succeed()

ROUTING_HEURISTIC = {
    "shortest_queue": shortest_queue_routing_heuristic,
    "random": random_routing_heuristic,
    "FIFO": FIFO_routing_heuristic,
}
"""
A dictionary of available routing heuristics.
"""


from abc import ABC, abstractmethod
import itertools
from typing import Callable, Iterator, List, Optional, Tuple, Union
from typing_extensions import deprecated

import numpy as np
from pydantic import BaseModel, ConfigDict, Field, PrivateAttr, field_validator

from prodsys.models.time_model_data import (
    FunctionTimeModelData,
    SampleTimeModelData,
    ScheduledTimeModelData,
    DistanceTimeModelData,
    SequentialTimeModelData,
    ManhattanDistanceTimeModelData,
)
from prodsys.util.statistical_functions import FUNCTION_DICT, FunctionTimeModelEnum


class TimeModel(ABC, BaseModel):
    """
    Abstract base class for time models.
    """
    @abstractmethod
    def get_next_time(
        self,
        origin: Optional[List[float]] = None, # TOOO: rework this with kwargs
        target: Optional[List[float]] = None,
    ) -> float:
        """
        Returns the next time of the time model.

        Args:
            origin (Optional[List[float]], optional): The origin of the product for a transport. Defaults to None.
            target (Optional[List[float]], optional): The target of the product for a transport. Defaults to None.

        Returns:
            float: The next time of the time model.
        """
        pass

    @abstractmethod
    def get_expected_time(
        self,
        origin: Optional[List[float]] = None,
        target: Optional[List[float]] = None,
    ) -> float:
        """
        Returns the expected time of the time model.

        Args:
            origin (Optional[List[float]], optional): The origin of the product for a transport. Defaults to None.
            target (Optional[List[float]], optional): The target of the product for a transport. Defaults to None.

        Returns:
            float: The expected time of the time model.
        """
        pass


class FunctionTimeModel(TimeModel):
    """
    Class for time models that are based on a function.

    Args:
        time_model_data (FunctionTimeModelData): The time model data object.
        statistics_buffer (List[float], optional): A buffer for the statistics. Defaults to [].
        distribution_function_object (Callable[[FunctionTimeModelData], List[float]], optional): The distribution function object. Defaults to FUNCTION_DICT[FunctionTimeModelEnum.Constant].
    """
    time_model_data: FunctionTimeModelData
    statistics_buffer: List[float] = []
    distribution_function_object: Callable[
        [FunctionTimeModelData], List[float]
    ] = FUNCTION_DICT[FunctionTimeModelEnum.Constant]

    @field_validator("distribution_function_object")
    def initialize_distribution_function(cls, v, values):
        return FUNCTION_DICT[values["time_model_data"].distribution_function]

    def get_next_time(
        self,
        origin: Optional[List[float]] = None,
        target: Optional[List[float]] = None,
    ) -> float:
        """
        Returns the next time for a time model based on a sample value of the distribution function.

        Returns:
            float: The next time of the time model.
        """
        try:
            value = self.statistics_buffer.pop()
            if value < 0:
                return 0.1
            return value
        except IndexError:
            self._fill_buffer()
            return self.get_next_time()

    def _fill_buffer(self):
        self.statistics_buffer = self.distribution_function_object(
            self.time_model_data
        )

    def get_expected_time(
        self,
        origin: Optional[List[float]] = None,
        target: Optional[List[float]] = None,
    ) -> float:
        """
        Returns the expected time for a time model based on the expected value of the distribution function.

        Returns:
            float: The expected time of the time model.
        """
        return self.time_model_data.location
    

class SampleTimeModel(TimeModel):
    """
    Class for time models that are based on a sample of values. A random value from the sample is returned.

    Args:
        time_model_data (SampleTimeModelData): The time model data object.
    """
    time_model_data: SampleTimeModelData

    def get_next_time(
        self,
        origin: Optional[List[float]] = None,
        target: Optional[List[float]] = None,
    ) -> float:
        """
        Returns the next time for a time model based on a sample value of the samples.

        Returns:
            float: The next time of the time model.
        """
        return np.random.choice(self.time_model_data.samples, 1)[0]

    def get_expected_time(
        self,
        origin: Optional[List[float]] = None,
        target: Optional[List[float]] = None,
    ) -> float:
        return sum(self.time_model_data.samples) / len(self.time_model_data.samples)
    

class ScheduledTimeModel(TimeModel):
    """
    Class for time models that are based on a schedule of values. A value from the schedule is returned based on the current schedule state.

    Args:
        time_model_data (ScheduledTimeModelData): The time model data object.
    """
    time_model_data: ScheduledTimeModelData

    _time_value_iterator: Iterator[float] = PrivateAttr()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._time_value_iterator = self._get_time_value_iterator()

    model_config=ConfigDict(arbitrary_types_allowed=True)


    def _get_time_value_iterator(self) -> Iterator[float]:
        """
        Returns an iterator for the time values of the schedule.

        Returns:
            Iterator[float]: The iterator for the time values of the schedule.
        """
        schedule = self.time_model_data.schedule
        if self.time_model_data.absolute:
            relative_schedule = [schedule[0]] + [schedule[i] - schedule[i - 1] for i in range(1, len(schedule))]
        else:
            relative_schedule = schedule
        if self.time_model_data.cyclic:
            return itertools.cycle(relative_schedule)
        return iter(relative_schedule)

    def get_next_time(
        self,
        origin: Optional[List[float]] = None,
        target: Optional[List[float]] = None,
    ) -> float:
        """
        Returns the next time for a time model based on a schedule value of the schedule.

        Returns:
            float: The next time of the time model.
        """
        try:
            return next(self._time_value_iterator)
        except StopIteration:
            return -1


    def get_expected_time(
        self,
        origin: Optional[List[float]] = None,
        target: Optional[List[float]] = None,
    ) -> float:
        """
        Returns the expected time for a time model based on the mean time difference value of the schedule.

        Args:
            origin (Optional[List[float]], optional): Not used. Defaults to None.
            target (Optional[List[float]], optional): Not used. Defaults to None.

        Returns:
            float: The expected time of the time model.
        """
        if self.time_model_data.absolute:
            schedule = self.time_model_data.schedule
            relative_schedule = [schedule[0]] + [schedule[i] - schedule[i - 1] for i in range(1, len(schedule))]
        else:
            relative_schedule = self.time_model_data.schedule
        return sum(relative_schedule) / len(relative_schedule)
    
class DistanceTimeModel(TimeModel):
    """
    Class for time models that are based on a distance between two points and time calculation based on reaction time and speed and distance metric.

    Args:
        time_model_data (DistanceTimeModelData): The time model data object.
    """
    time_model_data: DistanceTimeModelData

    def calculate_distance(self, origin: List[float], target: List[float]) -> float:
        """
        Calculate the distance between two points.

        Args:
            origin (List[float]): The origin point.
            target (List[float]): The target point.

        Returns:
            float: The distance between the two points.
        """
        if self.time_model_data.metric == "euclidean":
            return np.linalg.norm(np.array(origin) - np.array(target))
        elif self.time_model_data.metric == "manhattan":
            return np.sum(np.abs(np.array(origin) - np.array(target)))
        else:
            raise ValueError(f"Unknown distance metric: {self.time_model_data.metric}")

    def get_next_time(
        self,
        origin: Optional[List[float]] = None,
        target: Optional[List[float]] = None,
    ) -> float:
        """
        Returns the next time for a time model based on the distance between two points and time calculation based on reaction time and speed.

        Args:
            origin (Optional[List[float]], optional): The origin of the product for a transport. Defaults to None.
            target (Optional[List[float]], optional): The target of the product for a transport. Defaults to None.

        Returns:
            float: The next time of the time model.
        """
        if origin is None or target is None:
            raise ValueError("Origin and target must be defined for DistanceTimeModel")
        distance = self.calculate_distance(origin, target)
        return distance / self.time_model_data.speed + self.time_model_data.reaction_time

    def get_expected_time(
        self,
        origin: Optional[List[float]] = None,
        target: Optional[List[float]] = None,
    ) -> float:
        """
        Returns the expected time for a time model based on the manhattan distance between two points and time calculation based on reaction time and speed.

        Args:
            origin (Optional[List[float]], optional): The origin of the product for a transport. Defaults to None.
            target (Optional[List[float]], optional): The target of the product for a transport. Defaults to None.

        Returns:
            float: The expected time of the time model.
        """
        return self.get_next_time(origin, target)

@deprecated(
    "SequentialTimeModel is deprecated and will be removed in the next release. Use SampleTimeModel instead.",
    category=None
)
class SequentialTimeModel(TimeModel):
    """
    Class for time models that are based on a sequence of values. A random value from the sequence is returned.

    Args:
        time_model_data (SequentialTimeModelData): The time model data object.
    """
    time_model_data: SequentialTimeModelData

    def get_next_time(
        self,
        origin: Optional[List[float]] = None,
        target: Optional[List[float]] = None,
    ) -> float:
        """
        Returns the next time for a time model based on a sample value of the sequence.

        Returns:
            float: The next time of the time model.
        """
        return np.random.choice(self.time_model_data.sequence, 1)[0]

    def get_expected_time(
        self,
        origin: Optional[List[float]] = None,
        target: Optional[List[float]] = None,
    ) -> float:
        """
        Returns the expected time for a time model based on the mean value of the sequence.

        Args:
            origin (Optional[List[float]], optional): Not used. Defaults to None.
            target (Optional[List[float]], optional): Not used. Defaults to None.

        Returns:
            float: The expected time of the time model.
        """
        return sum(self.time_model_data.sequence) / len(self.time_model_data.sequence)


@deprecated(
    "ManhattanDistanceTimeModel is deprecated and will be removed in the next release. Use DistanceTimeModel instead.",
    category=None
)
class ManhattanDistanceTimeModel(TimeModel):
    """
    Class for time models that are based on the manhattan distance between two points and time calculation based on reaction time and speed.

    Args:
        time_model_data (ManhattanDistanceTimeModelData): The time model data object.
    """
    time_model_data: ManhattanDistanceTimeModelData
    def get_next_time(
        self,
        origin: Optional[List[float]] = None,
        target: Optional[List[float]] = None,
    ) -> float:
        """
        Returns the next time for a time model based on the manhattan distance between two points and time calculation based on reaction time and speed.

        Args:
            origin (Optional[List[float]], optional): The origin of the product for a transport. Defaults to None.
            target (Optional[List[float]], optional): The target of the product for a transport. Defaults to None.

        Returns:
            float: The next time of the time model.
        """
        if origin is None or target is None:
            raise ("Origin and target must be defined for ManhattanDistanceTimeModel")  # type: ignore
        x_distance = abs(origin[0] - target[0])
        y_distance = abs(origin[1] - target[1])
        return (
            x_distance + y_distance
        ) / self.time_model_data.speed + self.time_model_data.reaction_time

    def get_expected_time(
        self,
        origin: Optional[List[float]] = None,
        target: Optional[List[float]] = None,
    ) -> float:
        """
        Returns the expected time for a time model based on the manhattan distance between two points and time calculation based on reaction time and speed.

        Args:
            origin (Optional[List[float]], optional): The origin of the product for a transport. Defaults to None.
            target (Optional[List[float]], optional): The target of the product for a transport. Defaults to None.

        Returns:
            float: The expected time of the time model.
        """
        if origin is None or target is None:
            raise ("Origin and target must be defined for ManhattanDistanceTimeModel")  # type: ignore
        return self.get_next_time(origin, target)


TIME_MODEL = Union[SampleTimeModel, ScheduledTimeModel, DistanceTimeModel, SequentialTimeModel, ManhattanDistanceTimeModel, FunctionTimeModel]
"""
Union type for all time models.
"""
